import numpy as np
import pytest

from quam.components import *
from quam.components.channels import Channel, IQChannel, SingleChannel
from quam.core import *
from quam.utils.dataclass import get_dataclass_attr_annotations

try:
    from qm.exceptions import NoScopeFoundException
except ImportError:
    NoScopeFoundException = IndexError


def test_drag_gaussian_pulse():
    drag_pulse = pulses.DragGaussianPulse(
        amplitude=1, sigma=4, alpha=2, anharmonicity=200e6, length=20, axis_angle=0
    )

    assert drag_pulse.operation == "control"
    assert drag_pulse.length == 20
    assert drag_pulse.get_attrs() == {
        "id": None,
        "length": 20,
        "axis_angle": 0.0,
        "digital_marker": None,
        "amplitude": 1,
        "sigma": 4,
        "alpha": 2,
        "anharmonicity": 200000000.0,
        "detuning": 0.0,
        "subtracted": True,
    }

    waveform = drag_pulse.calculate_waveform()
    assert len(waveform) == 20
    assert isinstance(waveform, np.ndarray)
    assert np.iscomplexobj(waveform)


def test_drag_cosine_pulse():
    drag_pulse = pulses.DragCosinePulse(
        amplitude=1, alpha=2, anharmonicity=200e6, length=20, axis_angle=0
    )

    assert drag_pulse.operation == "control"
    assert drag_pulse.length == 20
    assert drag_pulse.get_attrs() == {
        "id": None,
        "length": 20,
        "axis_angle": 0.0,
        "digital_marker": None,
        "amplitude": 1,
        "alpha": 2,
        "anharmonicity": 200000000.0,
        "detuning": 0.0,
    }

    waveform = drag_pulse.calculate_waveform()
    assert len(waveform) == 20
    assert isinstance(waveform, np.ndarray)
    assert np.iscomplexobj(waveform)


def test_channel():
    channel = Channel()
    d = channel.to_dict(include_defaults=False)

    assert d == {
        "__class__": "quam.components.channels.Channel",
    }


def test_IQ_channel():
    IQ_channel = IQChannel(
        opx_output_I=0,
        opx_output_Q=1,
        intermediate_frequency=100e6,
        frequency_converter_up=FrequencyConverter(
            mixer=Mixer(), local_oscillator=LocalOscillator()
        ),
    )
    d = IQ_channel.to_dict(include_defaults=False)
    assert d == {
        "__class__": "quam.components.channels.IQChannel",
        "opx_output_I": 0,
        "opx_output_Q": 1,
        "intermediate_frequency": 100e6,
        "frequency_converter_up": {
            "__class__": "quam.components.hardware.FrequencyConverter",
            "mixer": {"__class__": "quam.components.hardware.Mixer"},
            "local_oscillator": {
                "__class__": "quam.components.hardware.LocalOscillator",
            },
        },
    }


def test_single_pulse_IQ_channel():
    IQ_channel = IQChannel(
        id="IQ",
        opx_output_I=0,
        opx_output_Q=1,
        intermediate_frequency=100e6,
        frequency_converter_up=FrequencyConverter(
            mixer=Mixer(), local_oscillator=LocalOscillator()
        ),
    )
    IQ_channel.operations["X180"] = pulses.GaussianPulse(
        length=16, amplitude=1, sigma=12
    )

    cfg = {"pulses": {}, "waveforms": {}}
    pulse = IQ_channel.operations["X180"]

    # axis_angle = None translates to all signal on I
    pulse.apply_to_config(cfg)

    pulse.axis_angle = 90
    pulse.apply_to_config(cfg)


def test_IQ_pulse_single_channel():
    single_channel = SingleChannel(
        id="single",
        opx_output=0,
    )
    single_channel.operations["X180"] = pulses.DragPulse(
        length=16,
        axis_angle=0,
        amplitude=1,
        sigma=12,
        alpha=0.1,
        anharmonicity=200e6,
    )

    cfg = {"pulses": {}, "waveforms": {}}
    pulse = single_channel.operations["X180"]

    with pytest.raises(ValueError) as exc_info:
        pulse.apply_to_config(cfg)
    error_message = "Waveform type 'IQ' not allowed for SingleChannel 'single'"
    assert str(exc_info.value) == error_message


def test_IQ_pulse_play_validate():
    single_channel = SingleChannel(
        id="single",
        opx_output=0,
    )

    with pytest.raises(KeyError):
        single_channel.play("X180")

    with pytest.raises(NoScopeFoundException):
        single_channel.play("X180", validate=False)

    single_channel.operations["X180"] = pulses.DragPulse(
        length=16,
        axis_angle=0,
        amplitude=1,
        sigma=12,
        alpha=0.1,
        anharmonicity=200e6,
    )

    with pytest.raises(NoScopeFoundException):
        single_channel.play("X180")


def test_pulse_parent_channel():
    pulse = pulses.SquarePulse(length=60, amplitude=0)
    assert pulse.parent is None
    assert pulse.channel is None

    channel = SingleChannel(id="single", opx_output=0)
    pulse.parent = channel
    assert pulse.parent is channel
    assert pulse.channel is channel


def test_pulse_parent_parent_channel():
    pulse = pulses.SquarePulse(length=60, amplitude=0)
    channel = SingleChannel(id="single", opx_output=0)
    channel.operations["pulse"] = pulse
    assert pulse.parent is channel.operations
    assert pulse.channel is channel


@quam_dataclass
class QuamTestPulseReferenced(QuamRoot):
    channel: SingleChannel


def test_pulses_referenced():
    channel = SingleChannel(id="single", opx_output=("con1", 1))
    machine = QuamTestPulseReferenced(channel=channel)

    pulse = pulses.SquarePulse(length=60, amplitude=0)
    channel.operations["pulse"] = pulse
    channel.operations["pulse_referenced"] = "#./pulse"

    assert (
        channel.operations["pulse_referenced"] == channel.operations["pulse"] == pulse
    )

    state = machine.to_dict(include_defaults=False)

    machine_loaded = QuamTestPulseReferenced.load(state)

    pulse_loaded = machine_loaded.channel.operations["pulse"]
    assert isinstance(pulse_loaded, pulses.SquarePulse)
    assert pulse_loaded.to_dict(include_defaults=False) == pulse.to_dict(include_defaults=False)

    assert machine_loaded.channel.operations["pulse_referenced"] == pulse_loaded
    assert (
        machine_loaded.channel.operations.get_raw_value("pulse_referenced")
        == "#./pulse"
    )


def test_pulse_attr_annotations():
    from quam.components import pulses

    attr_annotations = get_dataclass_attr_annotations(pulses.SquareReadoutPulse)

    assert list(attr_annotations["required"]) == ["length", "amplitude"]


def test_deprecated_drag_pulse():
    with pytest.warns(
        DeprecationWarning,
        match="DragPulse is deprecated. Use DragGaussianPulse instead.",
    ):
        pulses.DragPulse(
            axis_angle=0, amplitude=1, sigma=4, alpha=2, anharmonicity=200e6, length=20
        )


def test_pulse_play(mocker):
    channel = SingleChannel(id="single", opx_output=("con1", 1))
    pulse = pulses.SquarePulse(length=60, amplitude=0)
    channel.operations["pulse"] = pulse

    mock_play = mocker.patch("quam.components.channels.play")
    channel.play("pulse", duration=100)
    mock_play.assert_called_once_with(
        pulse="pulse",
        element="single",
        duration=100,
        condition=None,
        chirp=None,
        truncate=None,
        timestamp_stream=None,
        continue_chirp=False,
        target="",
    )


def test_pulse_play_no_channel(mocker):
    pulse = pulses.SquarePulse(length=60, amplitude=0)
    with pytest.raises(ValueError):
        pulse.play()


def test_arbitrary_waveform_iq_channel_list_conversion():
    """Test that arbitrary waveforms on IQ channels convert both I and Q to lists"""
    IQ_channel = IQChannel(
        id="IQ",
        opx_output_I=("con1", 1),
        opx_output_Q=("con1", 2),
        intermediate_frequency=100e6,
        frequency_converter_up=FrequencyConverter(
            mixer=Mixer(), local_oscillator=LocalOscillator()
        ),
    )

    # Create a Gaussian pulse that returns an arbitrary waveform (numpy array)
    gaussian_pulse = pulses.GaussianPulse(
        length=16, amplitude=1.0, sigma=4.0, axis_angle=None
    )
    IQ_channel.operations["gaussian"] = gaussian_pulse

    cfg = {"pulses": {}, "waveforms": {}}
    gaussian_pulse.apply_to_config(cfg)

    # Check that both I and Q waveforms are lists
    pulse_config = cfg["pulses"][gaussian_pulse.pulse_name]
    i_waveform_name = pulse_config["waveforms"]["I"]
    q_waveform_name = pulse_config["waveforms"]["Q"]

    i_waveform = cfg["waveforms"][i_waveform_name]["samples"]
    q_waveform = cfg["waveforms"][q_waveform_name]["samples"]

    assert isinstance(i_waveform, list), "I waveform should be a list"
    assert isinstance(q_waveform, list), "Q waveform should be a list"
    assert len(i_waveform) == 16, "I waveform should have correct length"
    assert len(q_waveform) == 16, "Q waveform should have correct length"
    # Q waveform should be all zeros since axis_angle=None
    assert all(q == 0.0 for q in q_waveform), "Q waveform should be all zeros"


def test_complex_arbitrary_waveform_iq_channel_list_conversion():
    """Test that complex arbitrary waveforms on IQ channels convert both I and Q to lists"""
    from quam.components.channels import IQChannel, MWChannel

    IQ_channel = IQChannel(
        id="IQ",
        opx_output_I=("con1", 1),
        opx_output_Q=("con1", 2),
        intermediate_frequency=100e6,
        frequency_converter_up=FrequencyConverter(
            mixer=Mixer(), local_oscillator=LocalOscillator()
        ),
    )

    # Create a custom pulse that returns a complex waveform
    @quam_dataclass
    class CustomComplexPulse(pulses.Pulse):
        amplitude: float = 1.0

        def waveform_function(self):
            # Return a complex numpy array
            return np.array([1 + 1j, 2 + 2j, 3 + 3j, 4 + 4j])

    complex_pulse = CustomComplexPulse(length=4)
    IQ_channel.operations["complex"] = complex_pulse

    cfg = {"pulses": {}, "waveforms": {}}
    complex_pulse.apply_to_config(cfg)

    # Check that both I and Q waveforms are lists
    pulse_config = cfg["pulses"][complex_pulse.pulse_name]
    i_waveform_name = pulse_config["waveforms"]["I"]
    q_waveform_name = pulse_config["waveforms"]["Q"]

    i_waveform = cfg["waveforms"][i_waveform_name]["samples"]
    q_waveform = cfg["waveforms"][q_waveform_name]["samples"]

    assert isinstance(i_waveform, list), "I waveform should be a list"
    assert isinstance(q_waveform, list), "Q waveform should be a list"
    assert len(i_waveform) == 4, "I waveform should have correct length"
    assert len(q_waveform) == 4, "Q waveform should have correct length"
    assert i_waveform == [1.0, 2.0, 3.0, 4.0], "I waveform should match real part"
    assert q_waveform == [1.0, 2.0, 3.0, 4.0], "Q waveform should match imaginary part"


def test_cosinebipolarpulse():
    # Basic instantiation and property checks
    pulse = pulses._CosineBipolarPulse(
        amplitude=1.0, flat_length=80, smoothing_length=20, post_zero_padding_length=10
    )
    assert (
        pulse.length == np.ceil((80 + 20 + 10) / 4) * 4
    )  # Length rounded to nearest multiple of 4
    assert pulse.amplitude == 1.0
    assert pulse.flat_length == 80
    assert pulse.axis_angle is None

    # Check waveform length and net-zero property
    waveform = np.array(pulse.waveform_function())
    assert (
        len(waveform) == np.ceil((80 + 20 + 10) / 4) * 4
    )  # Length rounded to nearest multiple of 4
    # Net-zero: sum should be close to zero
    print(np.sum(waveform.real))
    assert np.isclose(np.sum(waveform.real), 0, atol=1e-10)


def test_flattopgaussianpulse():
    # Basic instantiation and property checks
    pulse = pulses._FlatTopGaussianPulse(
        amplitude=1.0, flat_length=80, smoothing_length=20, post_zero_padding_length=10
    )
    assert (
        pulse.length == np.ceil((80 + 20 + 10) / 4) * 4
    )  # Length rounded to nearest multiple of 4
    assert pulse.amplitude == 1.0
    assert pulse.flat_length == 80
    assert pulse.axis_angle is None

    # Check waveform length
    waveform = np.array(pulse.waveform_function())
    assert (
        len(waveform) == np.ceil((80 + 20 + 10) / 4) * 4
    )  # Length rounded to nearest multiple of 4



from quam.utils.config import generate_config_final_actions


def test_config_no_overwrite_existing_offset():
    cfg = {
        "controllers": {
            "con1": {
                "analog_outputs": {1: {"offset": 0.5}},
                "analog_inputs": {2: {"offset": 0.5}},
            },
        },
    }

    generate_config_final_actions(cfg)

    assert cfg == {
        "controllers": {
            "con1": {
                "analog_outputs": {1: {"offset": 0.5}},
                "analog_inputs": {2: {"offset": 0.5}},
            },
        },
    }


def test_config_default_offset():
    cfg = {
        "controllers": {
            "con1": {
                "analog_outputs": {1: {}},
                "analog_inputs": {2: {}},
            },
        },
    }

    generate_config_final_actions(cfg)

    assert cfg == {
        "controllers": {
            "con1": {
                "analog_outputs": {1: {"offset": 0.0}},
                "analog_inputs": {2: {"offset": 0.0}},
            },
        },
    }


def test_config_default_offset_LF_FEM():
    cfg = {
        "controllers": {
            "con1": {
                "fems": {
                    2: {
                        "type": "LF",
                        "analog_outputs": {1: {}},
                        "analog_inputs": {2: {}},
                    },
                },
            },
        },
    }

    generate_config_final_actions(cfg)

    assert cfg == {
        "controllers": {
            "con1": {
                "fems": {
                    2: {
                        "type": "LF",
                        "analog_outputs": {1: {"offset": 0.0}},
                        "analog_inputs": {2: {"offset": 0.0}},
                    },
                },
            },
        },
    }


def test_config_no_overwrite_existing_offset_LF_FEM():
    cfg = {
        "controllers": {
            "con1": {
                "fems": {
                    2: {
                        "type": "LF",
                        "analog_outputs": {1: {"offset": 0.5}},
                        "analog_inputs": {2: {"offset": 0.5}},
                    },
                },
            },
        },
    }

    generate_config_final_actions(cfg)

    assert cfg == {
        "controllers": {
            "con1": {
                "fems": {
                    2: {
                        "type": "LF",
                        "analog_outputs": {1: {"offset": 0.5}},
                        "analog_inputs": {2: {"offset": 0.5}},
                    },
                },
            },
        },
    }


def test_config_default_offset_no_outputs_inputs_entries():
    cfg = {"controllers": {"con1": {}}}

    generate_config_final_actions(cfg)

    assert cfg == {"controllers": {"con1": {}}}


def test_config_default_offset_no_outputs_inputs():
    cfg = {
        "controllers": {
            "con1": {
                "analog_outputs": {},
                "analog_inputs": {},
            }
        }
    }

    generate_config_final_actions(cfg)

    assert cfg == {
        "controllers": {
            "con1": {
                "analog_outputs": {},
                "analog_inputs": {},
            }
        }
    }



from abc import ABC, abstractmethod
from typing import Any


__all__ = ["BaseMacro"]


class BaseMacro(ABC):
    """Base class for all macro types in the system"""

    @abstractmethod
    def apply(self, *args, **kwargs) -> Any:
        """Applies the macro operation"""
        pass



import pytest
from quam.components import *
from quam.components.ports.analog_inputs import OPXPlusAnalogInputPort
from quam.components.ports.analog_outputs import OPXPlusAnalogOutputPort


def test_in_out_single_channel_empty_error():
    with pytest.raises(TypeError):
        InOutSingleChannel()


def test_in_out_single_channel_no_id_error():
    channel = InOutSingleChannel(
        opx_output=("con1", 1),
        opx_input=("con1", 2),
    )
    with pytest.raises(AttributeError):
        channel.name


def test_in_out_single_channel():
    channel = InOutSingleChannel(
        id=1,
        opx_output=("con1", 1),
        opx_input=("con1", 2),
    )

    cfg = {"controllers": {}, "elements": {}}

    channel.apply_to_config(cfg)

    expected_cfg = {
        "controllers": {
            "con1": {
                "analog_inputs": {2: {"gain_db": 0, "shareable": False}},
                "analog_outputs": {1: {"delay": 0, "shareable": False}},
            }
        },
        "elements": {
            "ch1": {
                "operations": {},
                "outputs": {"out1": ("con1", 2)},
                "singleInput": {"port": ("con1", 1)},
                "smearing": 0,
                "time_of_flight": 140,
            }
        },
    }

    assert cfg == expected_cfg


def test_in_out_single_channel_ports():
    channel = InOutSingleChannel(
        id=1,
        opx_output=OPXPlusAnalogOutputPort("con1", 1),
        opx_input=OPXPlusAnalogInputPort("con1", 2),
    )

    cfg = {"controllers": {}, "elements": {}}

    channel.apply_to_config(cfg)
    channel.opx_output.apply_to_config(cfg)
    channel.opx_input.apply_to_config(cfg)

    expected_cfg = {
        "controllers": {
            "con1": {
                "analog_inputs": {2: {"gain_db": 0, "shareable": False}},
                "analog_outputs": {1: {"delay": 0, "shareable": False}},
            }
        },
        "elements": {
            "ch1": {
                "operations": {},
                "outputs": {"out1": ("con1", 2)},
                "singleInput": {"port": ("con1", 1)},
                "smearing": 0,
                "time_of_flight": 140,
            }
        },
    }

    assert cfg == expected_cfg



"""Tests for skip_save field metadata functionality.

This module tests the ability to exclude specific dataclass fields from
serialization using field(metadata={'skip_save': True}).
"""
from dataclasses import field
from typing import ClassVar
import tempfile
import json
from quam.core import QuamRoot, QuamComponent, QuamDict, QuamList, quam_dataclass


# Basic functionality tests
def test_skip_save_basic():
    """Test that fields with skip_save=True are excluded from to_dict()."""
    @quam_dataclass
    class TestComponent(QuamComponent):
        visible: int = 1
        hidden: int = field(default=2, metadata={"skip_save": True})

    obj = TestComponent()
    result = obj.to_dict()

    assert "visible" in result
    assert "hidden" not in result
    assert "__class__" in result
    assert result["visible"] == 1


def test_skip_save_false_explicit():
    """Test that fields with skip_save=False are included."""
    @quam_dataclass
    class TestComponent(QuamComponent):
        visible: int = 1
        also_visible: int = field(default=2, metadata={"skip_save": False})

    obj = TestComponent()
    result = obj.to_dict()

    assert "visible" in result
    assert "also_visible" in result
    assert result["also_visible"] == 2


def test_skip_save_no_metadata():
    """Test that fields without metadata are included by default."""
    @quam_dataclass
    class TestComponent(QuamComponent):
        field_a: int = 1
        field_b: int = 2

    obj = TestComponent()
    result = obj.to_dict()

    assert "field_a" in result
    assert "field_b" in result


def test_skip_save_multiple_fields():
    """Test multiple fields with mixed skip_save values."""
    @quam_dataclass
    class TestComponent(QuamComponent):
        visible_1: int = 1
        hidden_1: int = field(default=2, metadata={"skip_save": True})
        visible_2: str = "test"
        hidden_2: float = field(default=3.14, metadata={"skip_save": True})

    obj = TestComponent()
    result = obj.to_dict()

    assert "visible_1" in result
    assert "visible_2" in result
    assert "hidden_1" not in result
    assert "hidden_2" not in result


def test_skip_save_runtime_accessible():
    """Test that skipped fields remain accessible at runtime."""
    @quam_dataclass
    class TestComponent(QuamComponent):
        visible: int = 1
        hidden: int = field(default=2, metadata={"skip_save": True})

    obj = TestComponent()

    # Runtime access should work
    assert obj.visible == 1
    assert obj.hidden == 2

    # Modification should work
    obj.hidden = 99
    assert obj.hidden == 99

    # But still excluded from serialization
    result = obj.to_dict()
    assert "hidden" not in result


# Nested component tests
def test_skip_save_nested_component():
    """Test that skipped field with QuamComponent excludes entire subtree."""
    @quam_dataclass
    class InnerComponent(QuamComponent):
        inner_val: int = 1

    @quam_dataclass
    class OuterComponent(QuamComponent):
        visible: int = 2
        hidden: InnerComponent = field(
            default_factory=InnerComponent, metadata={"skip_save": True}
        )

    obj = OuterComponent()
    result = obj.to_dict()

    assert "visible" in result
    assert "hidden" not in result
    # Inner component's inner_val should not appear anywhere
    assert "inner_val" not in str(result)


def test_skip_save_deeply_nested():
    """Test skip_save with deeply nested component tree."""
    @quam_dataclass
    class Level3(QuamComponent):
        val3: int = 3

    @quam_dataclass
    class Level2(QuamComponent):
        val2: int = 2
        level3: Level3 = field(default_factory=Level3)

    @quam_dataclass
    class Level1(QuamComponent):
        val1: int = 1
        level2: Level2 = field(default_factory=Level2, metadata={"skip_save": True})

    obj = Level1()
    result = obj.to_dict()

    assert "val1" in result
    assert "level2" not in result
    # Entire subtree should be excluded
    assert "val2" not in str(result)
    assert "val3" not in str(result)


def test_skip_save_with_quam_dict_in_field():
    """Test skip_save on field containing QuamDict."""
    @quam_dataclass
    class TestComponent(QuamComponent):
        visible: int = 1
        hidden_dict: QuamDict = field(
            default_factory=QuamDict, metadata={"skip_save": True}
        )

    obj = TestComponent()
    obj.hidden_dict["key"] = "value"
    result = obj.to_dict()

    assert "visible" in result
    assert "hidden_dict" not in result


def test_skip_save_with_quam_list_in_field():
    """Test skip_save on field containing QuamList."""
    @quam_dataclass
    class TestComponent(QuamComponent):
        visible: int = 1
        hidden_list: QuamList = field(
            default_factory=QuamList, metadata={"skip_save": True}
        )

    obj = TestComponent()
    obj.hidden_list.append("item")
    result = obj.to_dict()

    assert "visible" in result
    assert "hidden_list" not in result


# Inheritance tests
def test_skip_save_inherited_field():
    """Test that inherited field with skip_save metadata is respected."""
    @quam_dataclass
    class ParentComponent(QuamComponent):
        inherited_visible: int = 1
        inherited_hidden: int = field(default=2, metadata={"skip_save": True})

    @quam_dataclass
    class ChildComponent(ParentComponent):
        child_field: int = 3

    obj = ChildComponent()
    result = obj.to_dict()

    assert "inherited_visible" in result
    assert "child_field" in result
    assert "inherited_hidden" not in result


def test_skip_save_field_override_removes_metadata():
    """Test that overriding a field in child class can remove skip_save."""
    @quam_dataclass
    class ParentComponent(QuamComponent):
        field: int = field(default=1, metadata={"skip_save": True})

    @quam_dataclass
    class ChildComponent(ParentComponent):
        field: int = 2  # Override without metadata

    obj = ChildComponent()
    result = obj.to_dict()

    # Field should now be visible since metadata was removed
    assert "field" in result
    assert result["field"] == 2


def test_skip_save_multiple_inheritance_levels():
    """Test skip_save behavior across multiple inheritance levels."""
    @quam_dataclass
    class GrandParent(QuamComponent):
        gp_visible: int = 1
        gp_hidden: int = field(default=2, metadata={"skip_save": True})

    @quam_dataclass
    class Parent(GrandParent):
        p_visible: int = 3
        p_hidden: int = field(default=4, metadata={"skip_save": True})

    @quam_dataclass
    class Child(Parent):
        c_visible: int = 5

    obj = Child()
    result = obj.to_dict()

    assert "gp_visible" in result
    assert "p_visible" in result
    assert "c_visible" in result
    assert "gp_hidden" not in result
    assert "p_hidden" not in result


# Integration tests
def test_skip_save_with_get_attrs():
    """Test that get_attrs() respects skip_save metadata."""
    @quam_dataclass
    class TestComponent(QuamComponent):
        visible: int = 1
        hidden: int = field(default=2, metadata={"skip_save": True})

    obj = TestComponent()
    attrs = obj.get_attrs()

    assert "visible" in attrs
    assert "hidden" not in attrs


def test_skip_save_save_load_roundtrip():
    """Test that skipped fields are excluded from saved JSON."""
    @quam_dataclass
    class TestRoot(QuamRoot):
        visible: int = 1
        hidden: int = field(default=2, metadata={"skip_save": True})

    obj = TestRoot()
    obj.hidden = 99  # Set to non-default value

    with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
        temp_path = f.name

    try:
        # Save to file
        obj.save(temp_path)

        # Load and inspect raw JSON
        with open(temp_path, "r") as f:
            saved_data = json.load(f)

        assert "visible" in saved_data
        assert "hidden" not in saved_data

        # Load from file
        loaded = TestRoot.load(temp_path)
        assert loaded.visible == 1
        # Hidden field should have default value since it wasn't saved
        assert loaded.hidden == 2

    finally:
        import os
        os.unlink(temp_path)


def test_skip_save_with_skip_attrs():
    """Test that skip_save works alongside _skip_attrs mechanism."""
    @quam_dataclass
    class TestComponent(QuamComponent):
        _skip_attrs = ["programmatic_skip"]

        visible: int = 1
        programmatic_skip: int = 2
        metadata_skip: int = field(default=3, metadata={"skip_save": True})

    obj = TestComponent()
    result = obj.to_dict()

    assert "visible" in result
    assert "programmatic_skip" not in result
    assert "metadata_skip" not in result


def test_skip_save_with_include_defaults_false():
    """Test skip_save interaction with include_defaults parameter."""
    @quam_dataclass
    class TestComponent(QuamComponent):
        visible_default: int = 1
        visible_nondefault: int = 2
        hidden_default: int = field(default=3, metadata={"skip_save": True})

    obj = TestComponent()
    obj.visible_nondefault = 99

    # With defaults
    result_with_defaults = obj.to_dict(include_defaults=True)
    assert "visible_default" in result_with_defaults
    assert "visible_nondefault" in result_with_defaults
    assert "hidden_default" not in result_with_defaults

    # Without defaults
    result_no_defaults = obj.to_dict(include_defaults=False)
    assert "visible_default" not in result_no_defaults
    assert "visible_nondefault" in result_no_defaults
    assert "hidden_default" not in result_no_defaults


def test_skip_save_with_multiple_metadata_entries():
    """Test that skip_save works alongside other metadata."""
    @quam_dataclass
    class TestComponent(QuamComponent):
        visible: int = 1
        hidden: int = field(
            default=2, metadata={"skip_save": True, "custom_key": "custom_value"}
        )

    obj = TestComponent()
    result = obj.to_dict()

    assert "visible" in result
    assert "hidden" not in result


def test_skip_save_nondefault_value():
    """Test that skip_save takes precedence over value being non-default."""
    @quam_dataclass
    class TestComponent(QuamComponent):
        hidden: int = field(default=1, metadata={"skip_save": True})

    obj = TestComponent(hidden=99)  # Set to non-default value
    result = obj.to_dict()

    # Should still be skipped even with non-default value
    assert "hidden" not in result


# Edge case tests
def test_skip_save_empty_metadata():
    """Test that empty metadata dict doesn't cause issues."""
    @quam_dataclass
    class TestComponent(QuamComponent):
        visible: int = field(default=1, metadata={})

    obj = TestComponent()
    result = obj.to_dict()

    assert "visible" in result


def test_skip_save_classvar_not_affected():
    """Test that ClassVar fields are not affected by skip_save logic."""
    @quam_dataclass
    class TestComponent(QuamComponent):
        class_var: ClassVar[int] = 99
        visible: int = 1
        hidden: int = field(default=2, metadata={"skip_save": True})

    obj = TestComponent()
    result = obj.to_dict()

    # ClassVar should not appear regardless
    assert "class_var" not in result
    assert "visible" in result
    assert "hidden" not in result


def test_skip_save_with_required_field():
    """Test skip_save on a required field (no default)."""
    @quam_dataclass
    class TestComponent(QuamComponent):
        visible: int
        hidden: int = field(metadata={"skip_save": True})

    obj = TestComponent(visible=1, hidden=2)
    result = obj.to_dict()

    assert "visible" in result
    assert "hidden" not in result


def test_skip_save_field_with_default_factory():
    """Test skip_save with default_factory fields."""
    @quam_dataclass
    class TestComponent(QuamComponent):
        visible: int = 1
        hidden: list = field(default_factory=list, metadata={"skip_save": True})

    obj = TestComponent()
    obj.hidden.append("item")
    result = obj.to_dict()

    assert "visible" in result
    assert "hidden" not in result



from typing import Dict
from dataclasses import field
import pytest

from quam.components.channels import (
    Channel,
    IQChannel,
    InOutIQChannel,
    InOutSingleChannel,
)

from quam.components.octave import Octave, OctaveUpConverter, OctaveDownConverter
from quam.core.quam_classes import QuamRoot, quam_dataclass


@quam_dataclass
class OctaveQuam(QuamRoot):
    octave: Octave
    channels: Dict[str, Channel] = field(default_factory=dict)


@pytest.fixture
def octave():
    return Octave(name="octave1")


def test_instantiate_octave(octave):
    assert octave.name == "octave1"
    assert octave.RF_outputs == {}
    assert octave.RF_inputs == {}
    assert octave.loopbacks == []


def test_empty_octave_config(octave, qua_config):
    machine = OctaveQuam(octave=octave)
    config = machine.generate_config()

    qua_config["octaves"] = {
        "octave1": {
            "RF_outputs": {},
            "RF_inputs": {},
            "IF_outputs": {},
            "loopbacks": [],
        }
    }

    assert config == qua_config


def test_empty_octave_empty_config(octave):
    cfg = {}
    octave.apply_to_config(config=cfg)

    expected_cfg = {
        "octaves": {
            "octave1": {
                "RF_outputs": {},
                "RF_inputs": {},
                "IF_outputs": {},
                "loopbacks": [],
            }
        }
    }
    assert cfg == expected_cfg


def test_octave_config_conflicting_entry(octave):
    machine = OctaveQuam(octave=octave)
    config = machine.generate_config()

    with pytest.raises(KeyError):
        octave.apply_to_config(config)


def test_frequency_converter_no_octave():
    converter = OctaveUpConverter(id=1, LO_frequency=2e9)
    assert converter.octave is None


def test_frequency_converter_octave(octave):
    converter = octave.RF_outputs[1] = OctaveUpConverter(id=1, LO_frequency=2e9)
    assert converter.octave is octave


def test_frequency_up_converter_apply_to_config(octave):
    converter = octave.RF_outputs[1] = OctaveUpConverter(id=1, LO_frequency=2e9)
    cfg = {}
    octave.apply_to_config(config=cfg)
    converter.apply_to_config(cfg)

    expected_cfg = {
        "octaves": {
            "octave1": {
                "RF_outputs": {
                    1: {
                        "LO_frequency": 2e9,
                        "LO_source": "internal",
                        "gain": 0,
                        "output_mode": "always_off",
                        "input_attenuators": "off",
                    }
                },
                "RF_inputs": {},
                "IF_outputs": {},
                "loopbacks": [],
            }
        }
    }
    assert cfg == expected_cfg


def test_frequency_down_converter_apply_to_config(octave):
    converter = octave.RF_inputs[1] = OctaveDownConverter(id=1, LO_frequency=2e9)
    cfg = {}
    octave.apply_to_config(config=cfg)
    converter.apply_to_config(cfg)

    expected_cfg = {
        "octaves": {
            "octave1": {
                "RF_outputs": {},
                "RF_inputs": {
                    1: {
                        "LO_frequency": 2e9,
                        "RF_source": "RF_in",
                        "LO_source": "internal",
                        "IF_mode_I": "direct",
                        "IF_mode_Q": "direct",
                    }
                },
                "IF_outputs": {},
                "loopbacks": [],
            }
        }
    }
    assert cfg == expected_cfg


def test_frequency_down_converter_with_IQchannel_apply_to_config(octave):
    channel = InOutIQChannel(
        opx_output_I=("con1", 3),
        opx_output_Q=("con1", 4),
        opx_input_I=("con1", 1),
        opx_input_Q=("con1", 2),
        frequency_converter_up=None,
        frequency_converter_down=None,
    )
    converter = octave.RF_inputs[1] = OctaveDownConverter(
        id=1, LO_frequency=2e9, channel=channel
    )
    cfg = {}
    octave.apply_to_config(config=cfg)
    converter.apply_to_config(cfg)

    expected_cfg = {
        "octaves": {
            "octave1": {
                "RF_outputs": {},
                "RF_inputs": {
                    1: {
                        "LO_frequency": 2e9,
                        "RF_source": "RF_in",
                        "LO_source": "internal",
                        "IF_mode_I": "direct",
                        "IF_mode_Q": "direct",
                    }
                },
                "IF_outputs": {
                    "IF_out1": {"port": ("con1", 1), "name": "out1"},
                    "IF_out2": {"port": ("con1", 2), "name": "out2"},
                },
                "loopbacks": [],
            }
        }
    }
    assert cfg == expected_cfg


def test_frequency_converter_down_existing_IF_outputs(octave):
    channel = InOutIQChannel(
        opx_output_I=("con1", 3),
        opx_output_Q=("con1", 4),
        opx_input_I=("con1", 2),
        opx_input_Q=("con1", 1),
        frequency_converter_up=None,
        frequency_converter_down=None,
    )
    converter = octave.RF_inputs[1] = OctaveDownConverter(
        id=1, LO_frequency=2e9, channel=channel
    )
    cfg = {}
    octave.apply_to_config(config=cfg)
    cfg["octaves"]["octave1"]["IF_outputs"] = {
        "IF_out1": {"port": ("con1", 1), "name": "out1"},
        "IF_out2": {"port": ("con1", 2), "name": "out2"},
    }

    with pytest.raises(ValueError):
        converter.apply_to_config(cfg)

    cfg = {}
    octave.apply_to_config(config=cfg)
    cfg["octaves"]["octave1"]["IF_outputs"] = {
        "IF_out1": {"port": ("con1", 2), "name": "out1"},
        "IF_out2": {"port": ("con1", 1), "name": "out2"},
    }

    converter.apply_to_config(cfg)

    assert cfg["octaves"]["octave1"]["IF_outputs"] == {
        "IF_out1": {"port": ("con1", 2), "name": "out1"},
        "IF_out2": {"port": ("con1", 1), "name": "out2"},
    }


def test_frequency_down_converter_with_single_channel_apply_to_config(octave):
    channel = InOutSingleChannel(
        opx_output=("con1", 3),
        opx_input=("con1", 3),
    )
    converter = octave.RF_inputs[1] = OctaveDownConverter(
        id=1, LO_frequency=2e9, channel=channel, IF_output_I=2
    )
    cfg = {}
    octave.apply_to_config(config=cfg)
    converter.apply_to_config(cfg)

    expected_cfg = {
        "octaves": {
            "octave1": {
                "RF_outputs": {},
                "RF_inputs": {
                    1: {
                        "LO_frequency": 2e9,
                        "RF_source": "RF_in",
                        "LO_source": "internal",
                        "IF_mode_I": "direct",
                        "IF_mode_Q": "direct",
                    }
                },
                "IF_outputs": {
                    "IF_out2": {"port": ("con1", 3), "name": "out1"},
                },
                "loopbacks": [],
            }
        }
    }
    assert cfg == expected_cfg


def test_instantiate_octave_default_connectivity(octave):
    octave.initialize_frequency_converters()

    assert list(octave.RF_outputs) == [1, 2, 3, 4, 5]
    for idx, RF_output in octave.RF_outputs.items():
        assert RF_output.octave == octave
        assert RF_output.id == idx

    assert list(octave.RF_inputs) == [1, 2]
    for idx, RF_input in octave.RF_inputs.items():
        assert RF_input.octave == octave
        assert RF_input.id == idx


def test_channel_add_RF_outputs(octave, qua_config):
    machine = OctaveQuam(octave=octave)
    octave.RF_outputs[2] = OctaveUpConverter(id=2, LO_frequency=2e9)

    channel = IQChannel(
        id="ch",
        opx_output_I=("con1", 1),
        opx_output_Q=("con1", 2),
        frequency_converter_up="#/octave/RF_outputs/2",
    )
    machine.channels["ch"] = channel
    channel.apply_to_config(qua_config)

    expected_cfg_elements = {
        "ch": {
            "RF_inputs": {"port": ("octave1", 2)},
            "operations": {},
        }
    }

    assert qua_config["elements"] == expected_cfg_elements


def test_channel_add_RF_inputs(octave, qua_config):
    machine = OctaveQuam(octave=octave)
    octave.RF_outputs[3] = OctaveUpConverter(id=3, LO_frequency=2e9)
    octave.RF_inputs[4] = OctaveDownConverter(id=4, LO_frequency=2e9)

    channel = InOutIQChannel(
        id="ch",
        opx_output_I=("con1", 1),
        opx_output_Q=("con1", 2),
        opx_input_I=("con1", 1),
        opx_input_Q=("con1", 2),
        frequency_converter_up="#/octave/RF_outputs/3",
        frequency_converter_down="#/octave/RF_inputs/4",
    )
    machine.channels["ch"] = channel
    channel.apply_to_config(qua_config)

    expected_cfg_elements = {
        "ch": {
            "RF_inputs": {"port": ("octave1", 3)},
            "RF_outputs": {"port": ("octave1", 4)},
            "operations": {},
            "smearing": 0,
            "time_of_flight": 140,
        }
    }

    assert qua_config["elements"] == expected_cfg_elements


def test_load_octave(octave):
    machine = OctaveQuam(octave=octave)
    octave.initialize_frequency_converters()

    d = machine.to_dict(include_defaults=False)

    d_expected = {
        "__class__": "test_octave.OctaveQuam",
        "octave": {
            "__class__": "quam.components.octave.Octave",
            "RF_inputs": {
                1: {"id": 1, "__class__": "quam.components.octave.OctaveDownConverter"},
                2: {
                    "id": 2,
                    "LO_source": "external",
                    "__class__": "quam.components.octave.OctaveDownConverter",
                },
            },
            "RF_outputs": {
                1: {"id": 1, "__class__": "quam.components.octave.OctaveUpConverter"},
                2: {"id": 2, "__class__": "quam.components.octave.OctaveUpConverter"},
                3: {"id": 3, "__class__": "quam.components.octave.OctaveUpConverter"},
                4: {"id": 4, "__class__": "quam.components.octave.OctaveUpConverter"},
                5: {"id": 5, "__class__": "quam.components.octave.OctaveUpConverter"},
            },
            "name": "octave1",
        },
    }
    assert d == d_expected

    machine2 = OctaveQuam.load(d)

    assert d == machine2.to_dict(include_defaults=False)


def test_frequency_converter_config_no_LO_frequency(octave):
    cfg = {}
    converter = octave.RF_outputs[1] = OctaveUpConverter(id=1)
    octave.apply_to_config(config=cfg)

    converter.apply_to_config(cfg)

    expected_cfg = {
        "octaves": {
            "octave1": {
                "RF_outputs": {},
                "RF_inputs": {},
                "IF_outputs": {},
                "loopbacks": [],
            }
        }
    }
    assert cfg == expected_cfg

    converter.channel = "something"

    with pytest.raises(ValueError):
        converter.apply_to_config(cfg)

    converter.channel = None
    converter.LO_frequency = 2e9

    converter.apply_to_config(cfg)

    expected_cfg = {
        "octaves": {
            "octave1": {
                "IF_outputs": {},
                "RF_inputs": {},
                "RF_outputs": {
                    1: {
                        "LO_frequency": 2000000000.0,
                        "LO_source": "internal",
                        "gain": 0,
                        "input_attenuators": "off",
                        "output_mode": "always_off",
                    }
                },
                "loopbacks": [],
            }
        }
    }
    assert cfg == expected_cfg



import pytest
from quam.components.channels import IQChannel, StickyChannelAddon
from quam.core.quam_classes import QuamRoot, quam_dataclass


def test_sticky_channel_no_duration_error():
    with pytest.raises(TypeError):
        StickyChannelAddon()

    StickyChannelAddon(duration=20)


@quam_dataclass
class SingleChannelQuam(QuamRoot):
    channel: IQChannel


def test_sticky_channel():
    channel = IQChannel(
        id="channel",
        opx_output_I=("con1", 1),
        opx_output_Q=("con1", 2),
        frequency_converter_up=None,
        sticky=StickyChannelAddon(duration=20),
    )
    machine = SingleChannelQuam(channel=channel)

    cfg = machine.generate_config()

    channel_cfg = cfg["elements"]["channel"]
    assert "sticky" in channel_cfg
    assert channel_cfg["sticky"] == {
        "duration": 20,
        "analog": True,
        "digital": True,
    }


def test_sticky_channel_disabled():
    channel = IQChannel(
        id="channel",
        opx_output_I=("con1", 1),
        opx_output_Q=("con1", 2),
        frequency_converter_up=None,
        sticky=StickyChannelAddon(duration=20, enabled=False),
    )
    machine = SingleChannelQuam(channel=channel)

    cfg = machine.generate_config()

    channel_cfg = cfg["elements"]["channel"]
    assert "sticky" not in channel_cfg



from .basic_quam import *
from .hardware import *
from .octave import *
from .channels import *
from . import pulses
from .quantum_components import *
from . import macro
from quam.config import get_quam_config

__all__ = [
    *basic_quam.__all__,
    *hardware.__all__,
    *channels.__all__,
    *octave.__all__,
    *quantum_components.__all__,
    "pulses",
    "macro",
    "get_quam_config",
]



from dataclasses import field
from typing import List, Optional
from quam.core.quam_classes import *


def test_basic_list_to_dict():
    l = QuamList([1, 2, 3])
    assert l.to_dict() == [1, 2, 3]


@quam_dataclass
class QuamTest(QuamComponent):
    int_val: int


def test_list_with_component_to_dict():
    c = QuamTest(int_val=42)
    quam_list = QuamList([c])
    assert quam_list.to_dict() == [
        {"int_val": 42, "__class__": "test_to_dict.QuamTest"}
    ]


def test_list_with_component_list_to_dict():
    l1 = QuamList([1, 2, 3])
    l2 = QuamList([1, 2, l1])
    assert l2.to_dict() == [1, 2, [1, 2, 3]]

    l3 = QuamList([1, 2, [4, 5, 6]])
    assert isinstance(l3[2], QuamList)
    assert l3.to_dict() == [1, 2, [4, 5, 6]]


def test_basic_dict_to_dict():
    d = QuamDict({"a": 1, "b": 2, "c": 3})
    assert d.to_dict() == {"a": 1, "b": 2, "c": 3}


def test_dict_with_component_to_dict():
    c = QuamTest(int_val=42)
    quam_dict = QuamDict({"a": c})
    assert quam_dict.to_dict() == {
        "a": {"int_val": 42, "__class__": "test_to_dict.QuamTest"}
    }


@quam_dataclass
class QuamBasicComponent(QuamComponent):
    a: int
    b: str


def test_quam_component_to_dict_basic():
    elem = QuamBasicComponent(a=42, b="foo")

    assert elem.to_dict() == {
        "a": 42,
        "b": "foo",
        "__class__": "test_to_dict.QuamBasicComponent",
    }


def test_quam_component_to_dict_nested():
    @quam_dataclass
    class QuamNestedComponent(QuamComponent):
        a: int
        b: str
        c: QuamComponent

    nested_elem = QuamBasicComponent(a=42, b="foo")
    elem = QuamNestedComponent(a=42, b="foo", c=nested_elem)

    assert elem.to_dict() == {
        "a": 42,
        "b": "foo",
        "c": {"a": 42, "b": "foo", "__class__": "test_to_dict.QuamBasicComponent"},
        "__class__": (
            "test_to_dict.test_quam_component_to_dict_nested.<locals>.QuamNestedComponent"
        ),
    }


def test_to_dict_nondefault():
    @quam_dataclass
    class QuamBasicComponent(QuamComponent):
        required_val: int
        optional_val: int = 42
        optional_list: List[int] = field(default_factory=list)

    quam_component = QuamBasicComponent(required_val=42)
    assert quam_component.to_dict(include_defaults=False) == {
        "required_val": 42,
        "__class__": "test_to_dict.test_to_dict_nondefault.<locals>.QuamBasicComponent",
    }
    assert quam_component.to_dict(include_defaults=True) == {
        "required_val": 42,
        "optional_val": 42,
        "optional_list": [],
        "__class__": "test_to_dict.test_to_dict_nondefault.<locals>.QuamBasicComponent",
    }

    quam_component = QuamBasicComponent(required_val=42, optional_val=43)
    assert quam_component.to_dict(include_defaults=False) == {
        "required_val": 42,
        "optional_val": 43,
        "__class__": "test_to_dict.test_to_dict_nondefault.<locals>.QuamBasicComponent",
    }
    assert quam_component.to_dict(include_defaults=True) == {
        "required_val": 42,
        "optional_val": 43,
        "optional_list": [],
        "__class__": "test_to_dict.test_to_dict_nondefault.<locals>.QuamBasicComponent",
    }

    quam_component = QuamBasicComponent(required_val=42, optional_list=["test"])
    assert quam_component.to_dict(include_defaults=False) == {
        "required_val": 42,
        "optional_list": ["test"],
        "__class__": "test_to_dict.test_to_dict_nondefault.<locals>.QuamBasicComponent",
    }
    assert quam_component.to_dict(include_defaults=True) == {
        "required_val": 42,
        "optional_val": 42,
        "optional_list": ["test"],
        "__class__": "test_to_dict.test_to_dict_nondefault.<locals>.QuamBasicComponent",
    }

    quam_component = QuamBasicComponent(required_val=42)
    quam_component.optional_list.append("test")
    assert quam_component.to_dict(include_defaults=False) == {
        "required_val": 42,
        "optional_list": ["test"],
        "__class__": "test_to_dict.test_to_dict_nondefault.<locals>.QuamBasicComponent",
    }
    assert quam_component.to_dict(include_defaults=True) == {
        "required_val": 42,
        "optional_val": 42,
        "optional_list": ["test"],
        "__class__": "test_to_dict.test_to_dict_nondefault.<locals>.QuamBasicComponent",
    }


def test_omit_default_dict_field():
    @quam_dataclass
    class QuamBasicComponent(QuamComponent):
        d: dict = field(default_factory=dict)

    quam_component = QuamBasicComponent()
    assert quam_component.to_dict(include_defaults=False) == {
        "__class__": (
            "test_to_dict.test_omit_default_dict_field.<locals>.QuamBasicComponent"
        ),
    }


def test_omit_default_list_field():
    @quam_dataclass
    class QuamBasicComponent(QuamComponent):
        l: list = field(default_factory=list)

    quam_component = QuamBasicComponent()
    assert quam_component.to_dict(include_defaults=False) == {
        "__class__": (
            "test_to_dict.test_omit_default_list_field.<locals>.QuamBasicComponent"
        ),
    }


def test_optional_list_to_dict():
    @quam_dataclass
    class QuamBasicComponent(QuamComponent):
        l: Optional[List[int]] = None

    quam_component = QuamBasicComponent()
    assert quam_component.to_dict(include_defaults=False) == {
        "__class__": (
            "test_to_dict.test_optional_list_to_dict.<locals>.QuamBasicComponent"
        ),
    }

    quam_component = QuamBasicComponent(l=[1, 2, 3])
    assert quam_component.to_dict() == {
        "l": [1, 2, 3],
        "__class__": (
            "test_to_dict.test_optional_list_to_dict.<locals>.QuamBasicComponent"
        ),
    }


def test_list_to_dict_nondefault():
    @quam_dataclass
    class QuamBasicComponent(QuamComponent):
        l: int = 42

    quam_component = QuamBasicComponent(l=[1, 2, 3])
    assert quam_component.to_dict() == {
        "l": [1, 2, 3],
        "__class__": (
            "test_to_dict.test_list_to_dict_nondefault.<locals>.QuamBasicComponent"
        ),
    }



def test_relative_list_referencing(BareQuamRoot):
    root = BareQuamRoot()
    root.l = [1, 2, 3, "#./0"]
    assert root.l[3] == 1

    root.l.append("#../")
    assert root.l[4] == root


def test_relative_dict_referencing(BareQuamRoot):
    root = BareQuamRoot()
    root.d = {"a": 1, "b": 2, "c": "#./a"}
    assert root.d.c == 1
    assert root.d["c"] == 1

    root.d["d"] = "#../"
    assert root.d.d == root
    assert root.d["d"] == root


def test_relative_component_referencing(BareQuamRoot, BareQuamComponent):
    root = BareQuamRoot()
    root.comp = BareQuamComponent()
    root.comp.a = 42
    root.comp.b = "#./a"

    assert root.comp.b == 42

    root.comp.c = "#../"
    assert root.comp.c == root

    root.comp.d = "#../comp"
    assert root.comp.d == root.comp



from typing import List

from quam.core import *


def test_base_quam_component_reference(BareQuamRoot, BareQuamComponent):
    quam_elem = BareQuamComponent()
    quam_elem.a = "#/test"
    assert quam_elem.a == "#/test"

    root = BareQuamRoot()
    root.elem = quam_elem
    assert quam_elem.a == "#/test"

    root.test = 42
    assert quam_elem.a == 42


@quam_dataclass
class QuamComponentTest(QuamComponent):
    int_val: int


def test_quam_component_reference_after_initialization(BareQuamRoot):
    quam_elem = QuamComponentTest(int_val=42)
    assert quam_elem.get_root() is None
    quam_elem.int_val = "#/test"
    assert quam_elem.int_val == "#/test"

    root = BareQuamRoot()
    root.elem = quam_elem
    assert quam_elem.int_val == "#/test"

    root.test = 42
    assert quam_elem.int_val == 42


def test_quam_component_reference_during_initialization(BareQuamRoot):
    quam_elem = QuamComponentTest(int_val="#/test")
    assert quam_elem.int_val == "#/test"

    root = BareQuamRoot()
    root.elem = quam_elem
    assert quam_elem.int_val == "#/test"

    root.test = 42
    assert quam_elem.int_val == 42


def test_basic_reference():
    @quam_dataclass
    class QuamRootTest(QuamRoot):
        quam_elem1: QuamComponentTest
        quam_elem2: QuamComponentTest

    quam_elem1 = QuamComponentTest(int_val=1)
    quam_elem2 = QuamComponentTest(int_val="#/quam_elem1/int_val")

    quam = QuamRootTest(quam_elem1=quam_elem1, quam_elem2=quam_elem2)

    assert quam_elem1.int_val == 1
    assert quam_elem2.int_val == 1

    quam_elem2.int_val = "#/quam_elem1"
    assert list(quam.iterate_components()) == [quam_elem1, quam_elem2]


def test_list_referencing():
    @quam_dataclass
    class QuamRootTest(QuamRoot):
        quam_elems: List[QuamComponentTest]
        quam_elem2: QuamComponentTest

    quam_elems = [QuamComponentTest(int_val=k) for k in range(5)]
    quam_elem2 = QuamComponentTest(int_val="#/quam_elems/3/int_val")

    quam = QuamRootTest(quam_elems=quam_elems, quam_elem2=quam_elem2)

    for k, elem in enumerate(quam.quam_elems):
        assert elem.int_val == k

    assert quam_elem2.int_val == 3


def test_reference_dict_elem():
    @quam_dataclass
    class QuamRootTest(QuamRoot):
        quam_elem_dict: dict
        quam_elem2: QuamComponentTest

    quam_elem_dict = QuamDict(port_I=2)
    quam_elem2 = QuamComponentTest(int_val="#/quam_elem_dict/port_I")

    QuamRootTest(quam_elem_dict=quam_elem_dict, quam_elem2=quam_elem2)

    assert quam_elem2.int_val == 2


# TODO Test referencing when a quam element is added to a quam



from typing import Optional
import pytest
from quam.components import Qubit
from quam.components.channels import IQChannel
from quam.components.pulses import SquarePulse
from quam.core.quam_classes import QuamRoot, quam_dataclass


def test_qubit_name_int():
    qubit = Qubit(id=0)
    assert qubit.name == "q0"


def test_qubit_name_str():
    qubit = Qubit(id="qubit0")
    assert qubit.name == "qubit0"


def test_qubit_channels(mock_qubit_with_resonator):
    assert mock_qubit_with_resonator.channels == {
        "xy": mock_qubit_with_resonator.xy,
        "resonator": mock_qubit_with_resonator.resonator,
    }


def test_qubit_channels_referenced(mock_qubit):
    # Set resonator as a reference to xy channel
    mock_qubit.resonator = "#./xy"

    assert mock_qubit.channels == {
        "xy": mock_qubit.xy,
        "resonator": mock_qubit.xy,
    }


def test_qubit_get_pulse_not_found(mock_qubit):
    with pytest.raises(ValueError, match="Pulse test_pulse not found"):
        mock_qubit.get_pulse("test_pulse")


def test_qubit_get_pulse_not_unique(mock_qubit_with_resonator):
    mock_qubit_with_resonator.xy.operations["test_pulse"] = SquarePulse(
        length=100, amplitude=1.0
    )
    mock_qubit_with_resonator.resonator.operations["test_pulse"] = SquarePulse(
        length=100, amplitude=1.0
    )

    with pytest.raises(ValueError, match="Pulse test_pulse is not unique"):
        mock_qubit_with_resonator.get_pulse("test_pulse")


def test_qubit_get_pulse_unique(mock_qubit):
    pulse = SquarePulse(length=100, amplitude=1.0)
    mock_qubit.xy.operations["test_pulse"] = pulse

    assert mock_qubit.get_pulse("test_pulse") == pulse


def test_qubit_align(mock_qubit_with_resonator, mock_qubit, mocker):
    mocker.patch("quam.components.quantum_components.qubit.align")
    mock_qubit_with_resonator.align(mock_qubit)

    from quam.components.quantum_components.qubit import align

    align.assert_called_once()
    called_args, _ = align.call_args
    assert set(called_args) == {"q1.xy", "q1.resonator", "q0.xy"}


def test_qubit_get_macros(mock_qubit):
    assert mock_qubit.macros == {}
    assert mock_qubit.get_macros() == {"align": mock_qubit.align}


def test_qubit_apply_align(mock_qubit_with_resonator, mocker):
    mocker.patch("quam.components.quantum_components.qubit.align")
    mock_qubit_with_resonator.align()

    from quam.components.quantum_components.qubit import align

    align.assert_called_once()
    called_args, _ = align.call_args
    assert set(called_args) == {"q1.xy", "q1.resonator"}


def test_qubit_inferred_id_direct():
    """Test inferred_id when id is a direct value"""
    qubit = Qubit(id=0)
    assert qubit.inferred_id == 0


def test_qubit_inferred_id_with_parent(test_quam):
    """Test inferred_id when id is a reference and qubit has parent"""
    test_quam.qubits["q2"] = Qubit()
    assert test_quam.qubits["q2"].inferred_id == "q2"


def test_qubit_inferred_id_no_parent():
    """Test inferred_id when id is a reference but qubit has no parent"""
    qubit = Qubit(id="#./inferred_id")
    with pytest.raises(
        AttributeError, match="Cannot infer id .* not attached to a parent"
    ):
        _ = qubit.inferred_id



from typing import Tuple
from qm.qua import QuaVariableType
from quam.components import Qubit, QubitPair
from quam.core import OperationsRegistry


operations_registry = OperationsRegistry()


@operations_registry.register_operation
def x(qubit: Qubit, **kwargs):
    pass


@operations_registry.register_operation
def y(qubit: Qubit, **kwargs):
    pass


@operations_registry.register_operation
def Rx(qubit: Qubit, angle: float, **kwargs):
    pass


@operations_registry.register_operation
def cz(qubit_pair: QubitPair, **kwargs):
    pass


@operations_registry.register_operation
def measure(qubit: Qubit, **kwargs) -> QuaVariableType:
    pass


@operations_registry.register_operation
def align(qubits: Tuple[Qubit, ...]):
    pass



from quam.utils import get_class_from_path


def test_get_transmon_from_class_path():
    transmon_path = "quam.examples.superconducting_qubits.components.Transmon"
    transmon_class = get_class_from_path(transmon_path)
    from quam.examples.superconducting_qubits.components import Transmon

    assert transmon_class == Transmon



from typing import Any
import pytest

from quam.core import quam_dataclass
from quam.utils import ReferenceClass


def test_instantiate_reference_class():
    ReferenceClass()


def test_set_non_reference_attribute():
    reference_obj = ReferenceClass()
    reference_obj.test = 42
    assert reference_obj.test == 42


def test_base_reference_class():
    reference_obj = ReferenceClass()

    with pytest.raises(NotImplementedError):
        reference_obj._is_reference("a")

    with pytest.raises(NotImplementedError):
        reference_obj._get_referenced_value("a")

    with pytest.raises(AttributeError):
        reference_obj.a


class SubReferenceClass(ReferenceClass):
    def _is_reference(self, attr: str) -> bool:
        return isinstance(attr, str) and attr.startswith("#")

    def _get_referenced_value(self, attr: str) -> Any:
        return attr[1:]


def test_set_reference_attribute():
    reference_obj = SubReferenceClass()
    reference_obj.a = "#b"
    assert reference_obj.a == "b"

    reference_obj.a = "#c"
    assert reference_obj.a == "c"

    reference_obj.a = 42
    assert reference_obj.a == 42


@quam_dataclass
class SubReferenceDataClass(SubReferenceClass):
    a: float = 42

    def __post_init__(self) -> None:
        super().__init__()


def test_reference_dataclass():
    sub_reference_class = SubReferenceDataClass()
    assert sub_reference_class.a == 42

    sub_reference_class.c = "#d"
    assert sub_reference_class.c == "d"

    sub_reference_class.a = "#b"
    assert sub_reference_class.a == "b"
    assert sub_reference_class.c == "d"



from typing import Dict, TYPE_CHECKING, Any
from dataclasses import field

from quam.core import quam_dataclass
from quam.components.quantum_components import QuantumComponent, Qubit
from quam.utils import string_reference as str_ref

if TYPE_CHECKING:
    from quam.components.macro import QubitPairMacro

    MacroType = QubitPairMacro
else:
    MacroType = Any


@quam_dataclass
class QubitPair(QuantumComponent):
    id: str = "#./name"
    qubit_control: Qubit
    qubit_target: Qubit
    macros: Dict[str, MacroType] = field(default_factory=dict)

    @property
    def name(self) -> str:
        if not str_ref.is_reference(self.get_raw_value("id")):
            return self.id
        else:
            return f"{self.qubit_control.name}@{self.qubit_target.name}"

    def align(self):
        """Aligns the execution of all channels of both qubits"""
        self.qubit_control.align(self.qubit_target)



from typing import Dict, List
from quam.core.quam_classes import *


@quam_dataclass
class QuamTest(QuamComponent):
    int_val: int
    str_val: str


def test_attr_type_nonexisting():
    test_quam = QuamTest(int_val=1, str_val="test")
    assert not test_quam._val_matches_attr_annotation("nonexisting", None)
    assert not test_quam._val_matches_attr_annotation("nonexisting", 123)


def test_attr_type_basic():
    test_quam = QuamTest(int_val=1, str_val="test")
    assert test_quam._val_matches_attr_annotation("int_val", 1)
    assert test_quam._val_matches_attr_annotation("str_val", "hi")
    assert not test_quam._val_matches_attr_annotation("int_val", "hi")
    assert not test_quam._val_matches_attr_annotation("str_val", 1)


@quam_dataclass
class QuamTest2(QuamComponent):
    int_val: int
    str_val: str
    l1: list
    l2: List[int]
    d1: dict
    d2: Dict[str, int]


def test_attr_type_dict():
    test_quam = QuamTest2(
        int_val=1, str_val="test", l1=[1, 2, 3], l2=[1, 2, 3], d1={"a": 1}, d2={"a": 1}
    )
    assert test_quam._val_matches_attr_annotation("d1", {})
    assert test_quam._val_matches_attr_annotation("d2", {})
    assert not test_quam._val_matches_attr_annotation("d1", 42)
    assert not test_quam._val_matches_attr_annotation("d2", 42)


def test_attr_type_list():
    test_quam = QuamTest2(
        int_val=1, str_val="test", l1=[1, 2, 3], l2=[1, 2, 3], d1={"a": 1}, d2={"a": 1}
    )
    assert test_quam._val_matches_attr_annotation("l1", [])
    assert test_quam._val_matches_attr_annotation("l2", [])
    assert not test_quam._val_matches_attr_annotation("l1", 42)
    assert not test_quam._val_matches_attr_annotation("l2", 42)



import pytest
from quam.components.channels import MWChannel
from quam.components.ports.analog_outputs import MWFEMAnalogOutputPort


def test_mw_channel():
    opx_output = MWFEMAnalogOutputPort(
        controller_id="con1", fem_id=1, port_id=1, band=1, upconverter_frequency=1e9
    )
    channel = MWChannel(opx_output=opx_output, upconverter=1, intermediate_frequency=0)

    assert channel.LO_frequency == 1e9
    assert channel.RF_frequency == 1e9

    channel.intermediate_frequency = 100e6
    assert channel.RF_frequency == 1.1e9


def test_mw_channel_upconverters_with_frequency():
    """Test successful access to upconverter frequency from upconverters dictionary."""
    opx_output = MWFEMAnalogOutputPort(
        controller_id="con1",
        fem_id=1,
        port_id=1,
        band=1,
        upconverters={1: {"frequency": 5e9}, 2: {"frequency": 6e9}},
    )
    channel = MWChannel(opx_output=opx_output, upconverter=1, intermediate_frequency=0)

    assert channel.upconverter_frequency == 5e9
    assert channel.LO_frequency == 5e9

    # Test with different upconverter
    channel.upconverter = 2
    assert channel.upconverter_frequency == 6e9


def test_mw_channel_missing_frequency_key():
    """Test error handling when frequency key is missing from upconverter config."""
    opx_output = MWFEMAnalogOutputPort(
        controller_id="con1",
        fem_id=1,
        port_id=1,
        band=1,
        upconverters={1: {"power": 10}},  # Missing frequency key
    )
    channel = MWChannel(opx_output=opx_output, upconverter=1, intermediate_frequency=0)

    with pytest.raises(
        ValueError, match="'frequency' key not found in upconverter 1 configuration"
    ):
        _ = channel.upconverter_frequency


def test_mw_channel_missing_upconverter():
    """Test error handling when upconverter ID is not found in upconverters dictionary."""
    opx_output = MWFEMAnalogOutputPort(
        controller_id="con1",
        fem_id=1,
        port_id=1,
        band=1,
        upconverters={1: {"frequency": 5e9}},  # Only upconverter 1 exists
    )
    channel = MWChannel(
        opx_output=opx_output, upconverter=3, intermediate_frequency=0
    )  # Request upconverter 3

    with pytest.raises(
        ValueError, match="Upconverter 3 not found in upconverters dictionary"
    ):
        _ = channel.upconverter_frequency



from typing import ClassVar, Dict, Optional, TypeVar, Union
from dataclasses import field
from quam.components.ports.base_ports import FEMPort
from quam.core import quam_dataclass, QuamComponent
from quam.core.quam_classes import QuamBase
from .analog_outputs import (
    OPXPlusAnalogOutputPort,
    LFFEMAnalogOutputPort,
    MWFEMAnalogOutputPort,
)
from .analog_inputs import (
    OPXPlusAnalogInputPort,
    LFFEMAnalogInputPort,
    MWFEMAnalogInputPort,
)
from .digital_outputs import OPXPlusDigitalOutputPort, FEMDigitalOutputPort
from .digital_inputs import OPXPlusDigitalInputPort

__all__ = ["OPXPlusPortsContainer", "FEMPortsContainer"]

OPXPlusPortTypes = Union[
    OPXPlusAnalogInputPort,
    OPXPlusAnalogOutputPort,
    OPXPlusDigitalOutputPort,
    OPXPlusDigitalInputPort,
]
FEMPortTypes = Union[
    LFFEMAnalogInputPort,
    LFFEMAnalogOutputPort,
    MWFEMAnalogInputPort,
    MWFEMAnalogOutputPort,
    FEMDigitalOutputPort,
]


@quam_dataclass
class OPXPlusPortsContainer(QuamComponent):
    analog_outputs: Dict[Union[str, int], Dict[int, OPXPlusAnalogOutputPort]] = field(
        default_factory=dict
    )
    analog_inputs: Dict[Union[str, int], Dict[int, OPXPlusAnalogInputPort]] = field(
        default_factory=dict
    )
    digital_outputs: Dict[Union[str, int], Dict[int, OPXPlusDigitalOutputPort]] = field(
        default_factory=dict
    )
    digital_inputs: Dict[Union[str, int], Dict[int, OPXPlusDigitalInputPort]] = field(
        default_factory=dict
    )

    def _get_port(
        self,
        controller_id: Union[str, int],
        port_id: int,
        port_type: str,
        create: bool = False,
        **kwargs,
    ):
        controllers = getattr(self, f"{port_type}s")

        try:
            return controllers[controller_id][port_id]
        except KeyError:
            if not create:
                raise KeyError(
                    f"Could not find existing {port_type} port: "
                    f"{port_type} ({controller_id}, {port_id}"
                )

        controllers.setdefault(controller_id, {})
        ports = controllers[controller_id]

        if port_type == "analog_output":
            ports[port_id] = OPXPlusAnalogOutputPort(controller_id, port_id, **kwargs)
        elif port_type == "analog_input":
            ports[port_id] = OPXPlusAnalogInputPort(controller_id, port_id, **kwargs)
        elif port_type == "digital_output":
            ports[port_id] = OPXPlusDigitalOutputPort(controller_id, port_id, **kwargs)
        elif port_type == "digital_input":
            ports[port_id] = OPXPlusDigitalInputPort(controller_id, port_id, **kwargs)
        else:
            raise ValueError(f"Invalid port type: {port_type}")

        return ports[port_id]

    def reference_to_port(
        self,
        port_reference: Union[QuamComponent, str],
        attr: Optional[str] = None,
        create=False,
    ) -> OPXPlusPortTypes:
        if isinstance(port_reference, QuamComponent):
            reference = port_reference.get_reference(attr=attr)
            if reference is None:
                raise ValueError("Cannot get port from reference {port_reference}")
            port_reference = reference
        elems = port_reference.split("/")
        port_type, controller_id, port_id = elems[-3:]

        port_type = port_type[:-1]
        if controller_id.isdigit():
            controller_id = int(controller_id)
        port_id = int(port_id)

        return self._get_port(controller_id, port_id, port_type, create=create)

    def get_analog_output(
        self,
        controller_id: Union[str, int],
        port_id: int,
        create: bool = False,
        **kwargs,
    ) -> OPXPlusAnalogOutputPort:
        return self._get_port(
            controller_id, port_id, port_type="analog_output", create=create, **kwargs
        )

    def get_analog_input(
        self,
        controller_id: Union[str, int],
        port_id: int,
        create: bool = False,
        **kwargs,
    ) -> OPXPlusAnalogInputPort:
        return self._get_port(
            controller_id, port_id, port_type="analog_input", create=create, **kwargs
        )

    def get_digital_output(
        self,
        controller_id: Union[str, int],
        port_id: int,
        create: bool = False,
        **kwargs,
    ) -> OPXPlusDigitalOutputPort:
        return self._get_port(
            controller_id, port_id, port_type="digital_output", create=create, **kwargs
        )

    def get_digital_input(
        self,
        controller_id: Union[str, int],
        port_id: int,
        create: bool = False,
        **kwargs,
    ) -> OPXPlusDigitalInputPort:
        return self._get_port(
            controller_id, port_id, port_type="digital_input", create=create, **kwargs
        )


@quam_dataclass
class FEMPortsContainer(QuamComponent):
    num_port_elems: ClassVar[int] = 3
    analog_outputs: Dict[
        Union[str, int], Dict[int, Dict[int, LFFEMAnalogOutputPort]]
    ] = field(default_factory=dict)
    analog_inputs: Dict[Union[str, int], Dict[int, Dict[int, LFFEMAnalogInputPort]]] = (
        field(default_factory=dict)
    )
    mw_outputs: Dict[Union[str, int], Dict[int, Dict[int, MWFEMAnalogOutputPort]]] = (
        field(default_factory=dict)
    )
    mw_inputs: Dict[Union[str, int], Dict[int, Dict[int, MWFEMAnalogInputPort]]] = (
        field(default_factory=dict)
    )
    digital_outputs: Dict[
        Union[str, int], Dict[int, Dict[int, FEMDigitalOutputPort]]
    ] = field(default_factory=dict)

    def _get_port(
        self,
        controller_id: Union[str, int],
        fem_id: int,
        port_id: int,
        port_type: str,
        create: bool = False,
        **kwargs,
    ):
        controllers = getattr(self, f"{port_type}s")

        try:
            return controllers[controller_id][fem_id][port_id]
        except KeyError:
            if not create:
                raise KeyError(
                    f"Could not find existing {port_type} port: "
                    f"{port_type} ({controller_id}, {fem_id}, {port_id}"
                )

        controllers.setdefault(controller_id, {})
        fems = controllers[controller_id]
        fems.setdefault(fem_id, {})
        ports = fems[fem_id]

        if port_type == "analog_output":
            ports[port_id] = LFFEMAnalogOutputPort(
                controller_id, fem_id, port_id, **kwargs
            )
        elif port_type == "analog_input":
            ports[port_id] = LFFEMAnalogInputPort(
                controller_id, fem_id, port_id, **kwargs
            )
        elif port_type == "mw_output":
            # Set default values in kwargs before passing to constructor
            # Safe to mutate since kwargs is used only once per port creation
            if "upconverter_frequency" not in kwargs and "upconverters" not in kwargs:
                kwargs["upconverter_frequency"] = 5e9
            if "band" not in kwargs:
                kwargs["band"] = 1
            ports[port_id] = MWFEMAnalogOutputPort(
                controller_id, fem_id, port_id, **kwargs
            )
        elif port_type == "mw_input":
            # Set default values in kwargs before passing to constructor
            # Safe to mutate since kwargs is used only once per port creation
            if "band" not in kwargs:
                kwargs["band"] = 1
            if "downconverter_frequency" not in kwargs:
                kwargs["downconverter_frequency"] = 5e9
            ports[port_id] = MWFEMAnalogInputPort(
                controller_id,
                fem_id,
                port_id,
                **kwargs,
            )
        elif port_type == "digital_output":
            ports[port_id] = FEMDigitalOutputPort(
                controller_id, fem_id, port_id, **kwargs
            )
        else:
            raise ValueError(f"Invalid port type: {port_type}")

        return ports[port_id]

    def reference_to_port(
        self,
        port_reference: Union[QuamComponent, str],
        attr: Optional[str] = None,
        create=False,
    ) -> FEMPortTypes:
        if isinstance(port_reference, QuamBase):
            reference = port_reference.get_reference(attr=attr)
            if reference is None:
                raise ValueError("Cannot get port from reference {port_reference}")
            port_reference = reference

        elems = port_reference.split("/")
        port_type, controller_id, fem_id, port_id = elems[-4:]

        port_type = port_type[:-1]
        if controller_id.isdigit():
            controller_id = int(controller_id)
        fem_id = int(fem_id)
        port_id = int(port_id)

        return self._get_port(controller_id, fem_id, port_id, port_type, create=create)

    def get_analog_output(
        self,
        controller_id: Union[str, int],
        fem_id: int,
        port_id: int,
        create: bool = False,
        **kwargs,
    ) -> LFFEMAnalogOutputPort:
        return self._get_port(
            controller_id,
            fem_id,
            port_id,
            port_type="analog_output",
            create=create,
            **kwargs,
        )

    def get_analog_input(
        self,
        controller_id: Union[str, int],
        fem_id: int,
        port_id: int,
        create: bool = False,
        **kwargs,
    ) -> LFFEMAnalogInputPort:
        return self._get_port(
            controller_id,
            fem_id,
            port_id,
            port_type="analog_input",
            create=create,
            **kwargs,
        )

    def get_mw_output(
        self,
        controller_id: Union[str, int],
        fem_id: int,
        port_id: int,
        create: bool = False,
        **kwargs,
    ) -> MWFEMAnalogOutputPort:
        return self._get_port(
            controller_id,
            fem_id,
            port_id,
            port_type="mw_output",
            create=create,
            **kwargs,
        )

    def get_mw_input(
        self,
        controller_id: Union[str, int],
        fem_id: int,
        port_id: int,
        create: bool = False,
        **kwargs,
    ) -> MWFEMAnalogInputPort:
        return self._get_port(
            controller_id,
            fem_id,
            port_id,
            port_type="mw_input",
            create=create,
            **kwargs,
        )

    def get_digital_output(
        self,
        controller_id: Union[str, int],
        fem_id: int,
        port_id: int,
        create: bool = False,
        **kwargs,
    ) -> FEMDigitalOutputPort:
        return self._get_port(
            controller_id,
            fem_id,
            port_id,
            port_type="digital_output",
            create=create,
            **kwargs,
        )



from collections import UserDict
import functools
from typing import Callable, Optional, TypeVar, Any

from quam.core.operation import Operation


__all__ = ["OperationsRegistry"]

T = TypeVar("T", bound=Callable[..., Any])


class OperationsRegistry(UserDict):
    """A registry to store and manage operations."""

    def register_operation(self, func: Optional[T]) -> T:
        """
        Register a function as an operation.

        This method stores the function in the operations dictionary and returns a
        wrapped version of the function that maintains the original function's
        signature and docstring.

        Args:
            func (callable): The function to register as an operation.

        Returns:
            callable: The wrapped function.
        """
        # Optionally add this later such that we can pass parameters to the decorator
        # if func is None:
        #     return functools.partial(self.register_operation)

        operation = Operation(func)
        operation = functools.update_wrapper(operation, func)

        self[func.__name__] = operation

        return operation  # type: ignore



import pytest
from typing import Dict
from dataclasses import field
from quam.core.quam_classes import *
from collections import UserDict


@quam_dataclass
class BareQuamComponent(QuamComponent):
    pass


def test_empty_quam_dict():
    quam_dict = QuamDict()
    assert isinstance(quam_dict, QuamBase)
    assert not isinstance(quam_dict, QuamComponent)
    assert isinstance(quam_dict, UserDict)

    assert quam_dict.data == {}
    assert quam_dict._get_attr_names() == []
    assert quam_dict.get_attrs() == {}
    assert quam_dict.to_dict() == {}

    assert quam_dict == QuamDict()
    assert quam_dict == {}

    with pytest.raises(AttributeError):
        quam_dict.nonexisting_attr
    with pytest.raises(KeyError):
        quam_dict["nonexisting_attr"]
    with pytest.raises(AttributeError):
        quam_dict.get_raw_value("nonexisting_attr")

    for val in [42, True, False, None]:
        assert not quam_dict._attr_val_is_default("nonexisting_attr", val)


def test_quam_dict_nonempty():
    quam_dict = QuamDict(val1=42)
    assert quam_dict.data == {"val1": 42}
    assert quam_dict._get_attr_names() == ["val1"]
    assert quam_dict.get_attrs() == {"val1": 42}
    assert quam_dict.to_dict() == {"val1": 42}

    assert quam_dict == QuamDict(val1=42)
    assert quam_dict == {"val1": 42}

    assert quam_dict.val1 == 42
    assert quam_dict["val1"] == 42
    assert quam_dict.get_raw_value("val1") == 42

    for val in [42, True, False, None]:
        assert not quam_dict._attr_val_is_default("val1", val)


def test_quam_dict_adding_elements():
    quam_dict = QuamDict()
    quam_dict["a"] = 42
    assert dict(quam_dict) == {"a": 42}
    assert quam_dict.data == {"a": 42}

    quam_dict["list"] = [1, 2, 3]
    assert dict(quam_dict) == {"a": 42, "list": [1, 2, 3]}
    assert isinstance(quam_dict["list"], QuamList)
    assert quam_dict.data == {"a": 42, "list": [1, 2, 3]}
    assert quam_dict.list == [1, 2, 3]

    quam_dict.pop("list")
    assert dict(quam_dict) == {"a": 42}


def test_quam_dict_setattr():
    quam_dict = QuamDict()

    quam_dict.val1 = 42
    assert quam_dict.val1 == 42
    assert quam_dict.data == {"val1": 42}
    assert quam_dict._get_attr_names() == ["val1"]
    assert quam_dict.get_attrs() == {"val1": 42}
    with pytest.raises(KeyError):
        quam_dict["val2"]
    with pytest.raises(AttributeError):
        quam_dict.get_raw_value("val2")


def test_quam_dict_setitem():
    quam_dict = QuamDict()

    quam_dict["val1"] = 42
    assert quam_dict.val1 == 42
    assert quam_dict.data == {"val1": 42}
    assert quam_dict._get_attr_names() == ["val1"]
    assert quam_dict.get_attrs() == {"val1": 42}
    assert quam_dict.to_dict() == {"val1": 42}
    assert quam_dict["val1"] == 42
    assert quam_dict.get_raw_value("val1") == 42


def test_inner_quam_dict():
    quam_dict = QuamDict(inner_dict={"val1": 42})
    assert isinstance(quam_dict.inner_dict, QuamDict)

    quam_dict = QuamDict(inner_dict={"inner_inner_dict": {"val1": 42}})
    assert isinstance(quam_dict.inner_dict, QuamDict)
    assert isinstance(quam_dict.inner_dict.inner_inner_dict, QuamDict)

    quam_dict = QuamDict()
    quam_dict.inner_dict = {"val1": 42}
    assert isinstance(quam_dict.inner_dict, QuamDict)


def test_dict_from_quam_root():
    @quam_dataclass
    class TestDictQuamRoot(QuamRoot):
        test_dict: dict = field(default_factory=dict)

    quam_root = TestDictQuamRoot()
    assert isinstance(quam_root.test_dict, QuamDict)


def test_dict_from_quam_component():
    @quam_dataclass
    class TestDictQuamComponent(QuamComponent):
        test_dict: dict = field(default_factory=dict)

    quam_component = TestDictQuamComponent()
    assert isinstance(quam_component.test_dict, QuamDict)

    quam_component = TestDictQuamComponent(test_dict={"bla": 42})
    assert isinstance(quam_component.test_dict, QuamDict)
    assert quam_component.test_dict.bla == 42


def test_quam_dict_getattr():
    quam_dict = QuamDict(val1=42)
    assert quam_dict.val1 == 42
    assert hasattr(quam_dict, "val1")
    assert "val1" in quam_dict


def test_quam_dict_get_attr_names():
    quam_dict = QuamDict(val1=42)
    assert quam_dict._get_attr_names() == ["val1"]


def test_val_matches_annotation():
    @quam_dataclass
    class TestQuamComponent(QuamComponent):
        val_dict: dict
        val_float: float
        val_Dict: Dict
        val_Dict_annotated: Dict[str, int]

    assert TestQuamComponent._val_matches_attr_annotation("val_dict", {})
    assert TestQuamComponent._val_matches_attr_annotation("val_dict", {1: 2})
    assert TestQuamComponent._val_matches_attr_annotation("val_dict", QuamDict())
    assert not TestQuamComponent._val_matches_attr_annotation("val_dict", 42)

    assert not TestQuamComponent._val_matches_attr_annotation("val_float", {})
    assert not TestQuamComponent._val_matches_attr_annotation("val_float", {1: 2})
    assert not TestQuamComponent._val_matches_attr_annotation("val_float", QuamDict())
    assert TestQuamComponent._val_matches_attr_annotation("val_float", 42.0)

    assert TestQuamComponent._val_matches_attr_annotation("val_Dict", {})
    assert TestQuamComponent._val_matches_attr_annotation("val_Dict", {1: 2})
    assert TestQuamComponent._val_matches_attr_annotation("val_Dict", QuamDict())
    assert not TestQuamComponent._val_matches_attr_annotation("val_Dict", 42)

    assert TestQuamComponent._val_matches_attr_annotation("val_Dict_annotated", {})
    assert TestQuamComponent._val_matches_attr_annotation("val_Dict_annotated", {1: 2})
    assert TestQuamComponent._val_matches_attr_annotation(
        "val_Dict_annotated", QuamDict()
    )
    assert not TestQuamComponent._val_matches_attr_annotation("val_Dict_annotated", 42)


def test_dict_parent(BareQuamComponent):
    quam_dict = QuamDict({"a": BareQuamComponent()})
    assert quam_dict["a"].parent == quam_dict

    quam_dict["b"] = BareQuamComponent()
    assert quam_dict["b"].parent == quam_dict

    quam_dict.c = BareQuamComponent()
    assert quam_dict.c.parent == quam_dict


def test_dict_nested():
    quam_dict = QuamDict(nested={"nested2": {"a": 42}})
    assert isinstance(quam_dict.nested, QuamDict)
    assert isinstance(quam_dict.nested.nested2, QuamDict)


def test_quam_dict_repr():
    quam_dict = QuamDict(val1=42, val2=43)
    assert repr(quam_dict) == "{'val1': 42, 'val2': 43}"


def test_dict_unreferenced_value():
    d = QuamDict(val1="#./val2", val2=42)
    assert d.val1 == 42
    assert d.val2 == 42
    assert d.get_raw_value("val1") == "#./val2"


def test_quam_dict_int_keys():
    quam_dict = QuamDict({1: 2})
    assert quam_dict.data == {1: 2}
    assert quam_dict[1] == 2
    quam_dict.pop(1)
    assert quam_dict.data == {}
    with pytest.raises(KeyError):
        quam_dict[1]


def test_quam_dict_get_attr_int():
    quam_dict = QuamDict({1: 2})
    assert quam_dict.get_attr_name(2) == 1


def test_quam_dict_print_summary():
    quam_dict = QuamDict({"a": "b", 1: 2})

    from contextlib import redirect_stdout
    import io

    f = io.StringIO()
    with redirect_stdout(f):
        quam_dict.print_summary()
    s = f.getvalue()
    assert s == 'QuamDict (parent unknown):\n  a: "b"\n  1: 2\n'


def test_quam_dict_to_dict_include_defaults():
    """Test that include_defaults kwarg is properly passed through QuamDict.to_dict()

    Reproduces issue #166: https://github.com/qua-platform/quam/issues/166
    The bug is that QuamDict.to_dict() doesn't pass kwargs to super().to_dict()
    """
    @quam_dataclass
    class MyComponent(QuamComponent):
        a: int = 5
        b: str = "hello"

    @quam_dataclass
    class TestRoot(QuamRoot):
        component: Dict[str, MyComponent]

    # Create a component with default values
    machine = TestRoot(component={"test": MyComponent()})

    # Without include_defaults, defaults should not be in the dict
    result_without_defaults = machine.to_dict(include_defaults=False)
    assert "a" not in result_without_defaults["component"]["test"]
    assert "b" not in result_without_defaults["component"]["test"]

    # With include_defaults=True, defaults should be included
    result_with_defaults = machine.to_dict(include_defaults=True)
    assert result_with_defaults["component"]["test"]["a"] == 5
    assert result_with_defaults["component"]["test"]["b"] == "hello"


def test_quam_dict_to_dict_follow_references():
    """Test that follow_references kwarg is properly passed through QuamDict.to_dict()

    Related to issue #166: https://github.com/qua-platform/quam/issues/166
    The key point is that the parameter gets passed to nested components in the dict.
    """
    @quam_dataclass
    class MyComponent(QuamComponent):
        ref_value: str = "#./target"
        target: int = 42

    @quam_dataclass
    class TestRoot(QuamRoot):
        component: Dict[str, MyComponent]

    # Create components with a reference
    machine = TestRoot(component={"test": MyComponent()})

    # The main goal is to ensure both parameters are passed without errors
    # Whether the reference is resolved depends on the parent-child hierarchy
    result_without_follow = machine.to_dict(follow_references=False)
    result_with_follow = machine.to_dict(follow_references=True)

    # Both should succeed without KeyError - the parameters were properly passed
    assert "component" in result_without_follow
    assert "test" in result_without_follow["component"]
    assert "component" in result_with_follow
    assert "test" in result_with_follow["component"]



from dataclasses import dataclass, is_dataclass, fields, field
import pytest

from quam.utils.dataclass import get_dataclass_attr_annotations
from quam.core import quam_dataclass


def test_dataclass_inheritance_error():
    @dataclass
    class BaseDataClass:
        b: int = 42

    with pytest.raises(TypeError):

        @dataclass
        class DerivedNondefaultDataClass(BaseDataClass):
            c: int

    @dataclass
    class DerivedDefaultDataClass(BaseDataClass):
        c: int = 41


def test_basic_quam_dataclass():
    @quam_dataclass
    class C:
        int_val: int

    assert is_dataclass(C)

    with pytest.raises(TypeError):
        c = C()

    c = C(int_val=1)
    assert is_dataclass(c)
    assert c.int_val == 1

    f = fields(c)
    assert len(f) == 1
    assert f[0].name == "int_val"


def test_quam_dataclass_with_default():
    @quam_dataclass
    class C:
        int_val: int
        int_val_optional: int = 42

    assert is_dataclass(C)
    with pytest.raises(TypeError):
        c = C()

    c = C(int_val=2)
    assert is_dataclass(c)
    assert c.int_val == 2
    assert c.int_val_optional == 42

    f = fields(c)
    assert len(f) == 2
    assert f[0].name == "int_val"
    assert f[1].name == "int_val_optional"


def test_dataclass_inheritance():
    @quam_dataclass
    class BaseClass:
        int_val: int

    @quam_dataclass
    class DerivedClass(BaseClass):
        int_val2: int

    with pytest.raises(TypeError):
        d = DerivedClass()
    with pytest.raises(TypeError):
        d = DerivedClass(42)

    d = DerivedClass(int_val=42, int_val2=43)

    f = fields(d)
    assert len(f) == 2
    assert f[0].name == "int_val"
    assert f[1].name == "int_val2"


def test_dataclass_inheritance_optional_derived():
    @quam_dataclass
    class BaseClass:
        int_val: int

    @quam_dataclass
    class DerivedClass(BaseClass):
        int_val2: int = 43

    with pytest.raises(TypeError):
        d = DerivedClass()
    with pytest.raises(TypeError):
        d = DerivedClass(int_val2=42)

    d = DerivedClass(int_val=42)

    assert d.int_val == 42
    assert d.int_val2 == 43
    f = fields(d)
    assert len(f) == 2
    assert f[0].name == "int_val"
    assert f[1].name == "int_val2"

    d = DerivedClass(int_val=42, int_val2=44)

    assert d.int_val == 42
    assert d.int_val2 == 44
    f = fields(d)
    assert len(f) == 2
    assert f[0].name == "int_val"
    assert f[1].name == "int_val2"


def test_dataclass_inheritance_optional_base():
    @quam_dataclass
    class BaseClass:
        int_val: int = 42

    @quam_dataclass
    class DerivedClass(BaseClass):
        int_val2: int
        int_val3: int = 44

    # TODO Raise error once REQUIRED is verified
    with pytest.raises(TypeError):
        d = DerivedClass()

    with pytest.raises(TypeError):
        d = DerivedClass(43)

    with pytest.raises(TypeError):
        d = DerivedClass(int_val=43)

    with pytest.raises(TypeError):
        d = DerivedClass(int_val3=43)

    d = DerivedClass(int_val=43, int_val2=45)
    assert d.int_val == 43
    assert d.int_val2 == 45
    assert d.int_val3 == 44

    d = DerivedClass(int_val=43, int_val2=45, int_val3=46)
    assert d.int_val == 43
    assert d.int_val2 == 45
    assert d.int_val3 == 46

    d = DerivedClass(int_val2=47)
    assert d.int_val == 42
    assert d.int_val2 == 47
    assert d.int_val3 == 44


def test_subsubclass_default_factory():
    """This bug was found quite late, and has been fixed.
    It only occurred with Python < 3.10.

    In this case, dataclass C has a keyword arg, and so class C2 cannot have args.
    patch_dataclass therefore gives the default REQUIRED to attr2
    Next, class C3 has the same attr but with a default_factory.
    Because it has a default_factory, it isn't included in C3.__dict__
    As a result, getattr(C3, "attr2") won't return 2, but instead REQUIRED.
    """

    @quam_dataclass
    class C:
        attr: int = 2

    @quam_dataclass
    class C2(C):
        attr2: int  # Adds REQUIRED default

    @quam_dataclass
    class C3(C2):
        attr2: int = field(default_factory=lambda: 2)

    attr_annotations = get_dataclass_attr_annotations(C3)
    assert attr_annotations == {
        "required": {},
        "optional": {"attr": int, "attr2": int},
        "allowed": {"attr": int, "attr2": int},
    }


def test_quam_dataclass_with_kw_only():
    @quam_dataclass(kw_only=True)
    class C:
        int_val: int
        int_val_optional: int = 42

    assert is_dataclass(C)
    with pytest.raises(TypeError):
        c = C()
    with pytest.raises(TypeError):
        c = C(int_val_optional=42)

    c = C(int_val=2)
    assert is_dataclass(c)
    assert c.int_val == 2
    assert c.int_val_optional == 42

    f = fields(c)
    assert len(f) == 2
    assert f[0].name == "int_val"
    assert f[1].name == "int_val_optional"


def test_quam_dataclass_optional_field():
    from quam.core import QuamComponent

    @quam_dataclass
    class RootClass(QuamComponent):
        optional_root_attr: int = None

    @quam_dataclass
    class DerivedClass(RootClass):
        pass

    from quam.utils.dataclass import get_dataclass_attr_annotations

    attr_annotations = get_dataclass_attr_annotations(DerivedClass)

    assert list(attr_annotations["required"]) == []



"""Test to_dict() explicit parameter behavior (backward compatibility)."""

from dataclasses import field
from typing import List

import pytest
from quam.core.quam_classes import QuamComponent, quam_dataclass
from quam.config.models.quam import QuamConfig
from unittest.mock import patch, MagicMock


@quam_dataclass
class SimpleComponent(QuamComponent):
    required_val: int
    optional_val: int = 42
    optional_str: str = "default_string"
    optional_list: List[int] = field(default_factory=list)


@quam_dataclass
class NestedComponent(QuamComponent):
    name: str
    simple: SimpleComponent
    count: int = 10


def test_to_dict_explicit_include_defaults_true():
    """Test that explicit include_defaults=True includes defaults."""
    comp = SimpleComponent(required_val=1)
    result = comp.to_dict(include_defaults=True)

    assert result["required_val"] == 1
    assert result["optional_val"] == 42
    assert result["optional_str"] == "default_string"
    assert result["optional_list"] == []


def test_to_dict_explicit_include_defaults_false():
    """Test that explicit include_defaults=False excludes defaults."""
    comp = SimpleComponent(required_val=1)
    result = comp.to_dict(include_defaults=False)

    assert result["required_val"] == 1
    assert "optional_val" not in result
    assert "optional_str" not in result
    assert "optional_list" not in result


def test_to_dict_default_includes_defaults():
    """Test that to_dict() without parameters includes defaults (v3 default behavior)."""
    comp = SimpleComponent(required_val=1)
    result = comp.to_dict()

    assert result["required_val"] == 1
    assert result["optional_val"] == 42
    assert result["optional_str"] == "default_string"
    assert result["optional_list"] == []


def test_to_dict_nested_respects_include_defaults_param():
    """Test that nested components respect include_defaults parameter."""
    simple = SimpleComponent(required_val=5, optional_val=100)
    nested = NestedComponent(name="test", simple=simple)

    result = nested.to_dict(include_defaults=True)

    assert result["name"] == "test"
    assert result["count"] == 10  # default value included
    assert "simple" in result

    # Nested component should also include defaults
    simple_result = result["simple"]
    assert simple_result["required_val"] == 5
    assert simple_result["optional_val"] == 100
    assert simple_result["optional_str"] == "default_string"
    assert simple_result["optional_list"] == []


def test_to_dict_nested_includes_defaults_by_default():
    """Test that nested components include defaults by default (v3 default behavior)."""
    simple = SimpleComponent(required_val=5, optional_val=100)
    nested = NestedComponent(name="test", simple=simple)

    result = nested.to_dict()

    assert result["name"] == "test"
    assert result["count"] == 10  # default value included by default
    assert "simple" in result

    # Nested component should also include defaults
    simple_result = result["simple"]
    assert simple_result["required_val"] == 5
    assert simple_result["optional_val"] == 100
    assert simple_result["optional_str"] == "default_string"
    assert simple_result["optional_list"] == []



from quam.components.channels import InIQOutSingleChannel
from quam.components.ports.analog_inputs import OPXPlusAnalogInputPort
from quam.components.ports.analog_outputs import OPXPlusAnalogOutputPort
from quam.utils.dataclass import get_dataclass_attr_annotations


def test_in_single_channel_attr_annotations():
    attr_annotations = get_dataclass_attr_annotations(InIQOutSingleChannel)
    assert set(attr_annotations["required"]) == {
        "opx_output",
        "opx_input_I",
        "opx_input_Q",
    }
    assert set(attr_annotations["optional"]) == {
        "sticky",
        "operations",
        "filter_fir_taps",
        "filter_iir_taps",
        "intermediate_frequency",
        "opx_output_offset",
        "id",
        "digital_outputs",
        "time_of_flight",
        "smearing",
        "input_gain",
        "opx_input_offset_I",
        "opx_input_offset_Q",
        "frequency_converter_down",
        "thread",
        "core",
    }


def test_generate_config(qua_config):
    channel = InIQOutSingleChannel(
        id="in_out_channel",
        opx_output=("con1", 1),
        opx_input_I=("con1", 1),
        opx_input_Q=("con1", 2),
        frequency_converter_down=None,
    )

    channel.apply_to_config(qua_config)

    assert qua_config["controllers"] == {
        "con1": {
            "analog_inputs": {
                1: {"gain_db": 0, "shareable": False},
                2: {"gain_db": 0, "shareable": False},
            },
            "analog_outputs": {1: {"delay": 0, "shareable": False}},
        }
    }

    assert qua_config["elements"] == {
        "in_out_channel": {
            "operations": {},
            "outputs": {"out1": ("con1", 1), "out2": ("con1", 2)},
            "singleInput": {"port": ("con1", 1)},
            "smearing": 0,
            "time_of_flight": 140,
        }
    }


def test_generate_config_ports(qua_config):
    channel = InIQOutSingleChannel(
        id="in_out_channel",
        opx_output=OPXPlusAnalogOutputPort("con1", 1),
        opx_input_I=OPXPlusAnalogInputPort("con1", 1),
        opx_input_Q=OPXPlusAnalogInputPort("con1", 2),
        frequency_converter_down=None,
    )

    channel.apply_to_config(qua_config)
    channel.opx_output.apply_to_config(qua_config)
    channel.opx_input_I.apply_to_config(qua_config)
    channel.opx_input_Q.apply_to_config(qua_config)

    assert qua_config["controllers"] == {
        "con1": {
            "analog_inputs": {
                1: {"gain_db": 0, "shareable": False},
                2: {"gain_db": 0, "shareable": False},
            },
            "analog_outputs": {1: {"delay": 0, "shareable": False}},
        }
    }

    assert qua_config["elements"] == {
        "in_out_channel": {
            "operations": {},
            "outputs": {"out1": ("con1", 1), "out2": ("con1", 2)},
            "singleInput": {"port": ("con1", 1)},
            "smearing": 0,
            "time_of_flight": 140,
        }
    }



from typing import List

import pytest

from quam.core import *


def test_root_reference(BareQuamRoot):
    root = BareQuamRoot()
    assert root.get_root() is root

    assert root.get_reference() == "#/"


@quam_dataclass
class QuamComponentTest(QuamComponent):
    test_str: str


@quam_dataclass
class QuamRootTest(QuamRoot):
    quam_elem: QuamComponentTest
    quam_elem_list: List[QuamComponentTest]


def test_quam_component_no_parent_reference_error():
    quam_elem = QuamComponentTest(test_str="hi")
    with pytest.raises(AttributeError):
        quam_elem.get_reference()


def test_quam_component_reference():
    root = QuamRootTest(quam_elem=QuamComponentTest(test_str="hi"), quam_elem_list=[])

    assert root.get_reference() == "#/"
    assert root.quam_elem.get_reference() == "#/quam_elem"


def test_quam_list_reference():
    root = QuamRootTest(
        quam_elem=QuamComponentTest(test_str="hi"),
        quam_elem_list=[
            QuamComponentTest(test_str="hi"),
            QuamComponentTest(test_str=QuamComponentTest(test_str="hi")),
        ],
    )

    assert root.get_reference() == "#/"
    assert root.quam_elem.get_reference() == "#/quam_elem"
    assert root.quam_elem_list[0].get_reference() == "#/quam_elem_list/0"
    assert root.quam_elem_list[1].get_reference() == "#/quam_elem_list/1"
    assert (
        root.quam_elem_list[1].test_str.get_reference() == "#/quam_elem_list/1/test_str"
    )


def test_get_reference_attr():
    component = QuamComponentTest(test_str="hi")
    root = QuamRootTest(quam_elem=component, quam_elem_list=[])
    assert component.get_reference() == "#/quam_elem"
    assert component.get_reference("test_str") == "#/quam_elem/test_str"



from importlib.metadata import version, PackageNotFoundError

try:
    # Reads version from installed package metadata.
    __version__ = version("quam")
except PackageNotFoundError:
    # Fallback when package is not installed (e.g., running from source)
    __version__ = "0.0.0-unknown"



from copy import deepcopy
from quam.components import Channel, DigitalOutputChannel, pulses
from quam.components.ports.digital_outputs import OPXPlusDigitalOutputPort
from quam.core import QuamRoot, quam_dataclass
from quam.core.quam_instantiation import instantiate_quam_class


@quam_dataclass
class QuamTest(QuamRoot):
    channel: Channel


def test_digital_only_channel(qua_config):
    channel = Channel(
        id="channel",
        digital_outputs={"1": DigitalOutputChannel(opx_output=("con1", 1))},
    )

    quam = QuamTest(channel=channel)
    cfg = quam.generate_config()

    qua_config["controllers"] = {
        "con1": {"digital_outputs": {1: {"inverted": False, "shareable": False}}}
    }
    qua_config["elements"] = {
        "channel": {"digitalInputs": {"1": {"port": ("con1", 1)}}, "operations": {}}
    }

    assert cfg == qua_config


def test_digital_only_channel_with_port(qua_config):

    channel = Channel(
        id="channel",
        digital_outputs={
            "1": DigitalOutputChannel(
                opx_output=OPXPlusDigitalOutputPort(
                    "con1", 2, inverted=True, shareable=True
                )
            )
        },
    )

    quam = QuamTest(channel=channel)
    cfg = quam.generate_config()

    qua_config["controllers"] = {
        "con1": {"digital_outputs": {2: {"inverted": True, "shareable": True}}}
    }
    qua_config["elements"] = {
        "channel": {"digitalInputs": {"1": {"port": ("con1", 2)}}, "operations": {}}
    }

    assert cfg == qua_config


def test_digital_only_pulse(qua_config):
    channel = Channel(
        id="channel",
        operations={
            "digital": pulses.Pulse(length=100, digital_marker=[(1, 20, 0, 10)])
        },
        # digital_outputs={"1": DigitalOutputChannel(opx_output=("con1", 1))},
    )

    quam = QuamTest(channel=channel)
    cfg = quam.generate_config()

    qua_config["elements"] = {
        "channel": {"operations": {"digital": "channel.digital.pulse"}}
    }
    qua_config["pulses"]["channel.digital.pulse"] = {
        "length": 100,
        "operation": "control",
        "digital_marker": "channel.digital.dm",
    }
    qua_config["digital_waveforms"]["channel.digital.dm"] = {
        "samples": [(1, 20, 0, 10)]
    }

    assert cfg == qua_config


def test_instantiate_digital_channel():
    channel = instantiate_quam_class(
        quam_class=DigitalOutputChannel, contents={"opx_output": ["con1", 1]}
    )

    assert channel.opx_output == ("con1", 1)



import pytest
from typing import List, Dict, Literal

from quam.utils.general import validate_obj_type


def test_validate_standard_types():
    validate_obj_type(1, int)
    validate_obj_type(1.0, float)
    validate_obj_type("hello", str)
    validate_obj_type(":reference", str)
    validate_obj_type([1, 2, 3], list)
    validate_obj_type((1, 2, 3), tuple)
    validate_obj_type({"a": 1, "b": 2}, dict)
    validate_obj_type(True, bool)
    validate_obj_type(None, type(None))

    with pytest.raises(TypeError):
        validate_obj_type(1, str)
    with pytest.raises(TypeError):
        validate_obj_type("hello", int)


def test_validate_type_exceptions():
    validate_obj_type("#/reference", int)
    validate_obj_type("#/reference", str)
    validate_obj_type("#./reference", int)
    validate_obj_type("#./reference", str)
    validate_obj_type("#../reference", int)
    validate_obj_type("#../reference", str)

    validate_obj_type(None, int)
    validate_obj_type(None, str)


def test_validate_typing_list():
    validate_obj_type([1, 2, 3], List[int])
    with pytest.raises(TypeError):
        validate_obj_type([1, 2, 3], List[str])

    validate_obj_type([1, 2, 3], List)
    validate_obj_type(["a", "b", "c"], List)
    validate_obj_type(["a", "b", "c"], List[str])
    with pytest.raises(TypeError):
        validate_obj_type(["a", "b", "c"], List[int])


def test_validate_typing_dict():
    validate_obj_type({"a": 1, "b": 2}, dict)
    validate_obj_type({"a": 1, "b": 2}, Dict[str, int])
    with pytest.raises(TypeError):
        validate_obj_type({"a": 1, "b": 2}, Dict[str, str])

    validate_obj_type("#/reference", Dict[str, int])
    validate_obj_type("#./reference", Dict[str, int])
    validate_obj_type("#../reference", Dict[str, int])


def test_validate_typing_literal():
    validate_obj_type("a", Literal["a", "b", "c"])
    validate_obj_type("b", Literal["a", "b", "c"])

    with pytest.raises(TypeError):
        validate_obj_type("d", Literal["a", "b", "c"])

    with pytest.raises(TypeError):
        validate_obj_type(123, Literal["b", "c"])



import pytest

from quam.components.hardware import FrequencyConverter, BaseFrequencyConverter
from quam.core.quam_instantiation import instantiate_quam_class
from quam.core.deprecations import DeprecatedFrequencyConverterInstantiation


def test_deprecation_frequency_converter():
    assert not DeprecatedFrequencyConverterInstantiation.match(FrequencyConverter, {})
    assert DeprecatedFrequencyConverterInstantiation.match(BaseFrequencyConverter, {})
    assert not DeprecatedFrequencyConverterInstantiation.match(
        BaseFrequencyConverter,
        {"__class__": "quam.components.hardware.FrequencyConverter"},
    )

    cls, contents = DeprecatedFrequencyConverterInstantiation.apply(
        BaseFrequencyConverter, {}
    )
    assert cls == FrequencyConverter
    assert contents == {"__class__": "quam.components.hardware.FrequencyConverter"}

    cls, contents = DeprecatedFrequencyConverterInstantiation.apply(
        BaseFrequencyConverter,
        {"__class__": "quam.components.hardware.BaseFrequencyConverter"},
    )
    assert cls == FrequencyConverter
    assert contents == {"__class__": "quam.components.hardware.FrequencyConverter"}


def test_instantiate_frequency_converter_deprecation():
    contents = {}
    with pytest.deprecated_call():
        obj = instantiate_quam_class(BaseFrequencyConverter, contents)

    assert isinstance(obj, FrequencyConverter)
    assert contents == {"__class__": "quam.components.hardware.FrequencyConverter"}



import pytest
from quam.components import *
from quam.components.ports.analog_inputs import OPXPlusAnalogInputPort
from quam.components.ports.analog_outputs import OPXPlusAnalogOutputPort


def test_empty_in_out_IQ_channel():
    readout_resonator = InOutIQChannel(
        frequency_converter_up=FrequencyConverter(
            mixer=Mixer(), local_oscillator=LocalOscillator()
        ),
        opx_output_I=("con1", 1),
        opx_output_Q=("con1", 2),
        opx_input_I=("con1", 3),
        opx_input_Q=("con1", 4),
        intermediate_frequency=100e6,
    )

    assert isinstance(readout_resonator.frequency_converter_up.mixer, Mixer)

    assert isinstance(readout_resonator.local_oscillator, LocalOscillator)
    assert (
        readout_resonator.frequency_converter_up.local_oscillator
        is readout_resonator.local_oscillator
    )

    mixer = readout_resonator.frequency_converter_up.mixer
    assert mixer.intermediate_frequency == 100e6

    assert mixer.local_oscillator_frequency is None
    readout_resonator.local_oscillator.frequency = 5e9
    assert mixer.local_oscillator_frequency == 5e9

    with pytest.raises(AttributeError):
        mixer.name
    assert readout_resonator.id is None
    with pytest.raises(AttributeError):
        readout_resonator.name

    readout_resonator.id = 1

    d = readout_resonator.to_dict(include_defaults=False)
    assert d == {
        "frequency_converter_up": {
            "__class__": "quam.components.hardware.FrequencyConverter",
            "mixer": {"__class__": "quam.components.hardware.Mixer"},
            "local_oscillator": {
                "frequency": 5000000000.0,
                "__class__": "quam.components.hardware.LocalOscillator",
            },
        },
        "opx_output_I": ("con1", 1),
        "opx_output_Q": ("con1", 2),
        "opx_input_I": ("con1", 3),
        "opx_input_Q": ("con1", 4),
        "intermediate_frequency": 100000000.0,
        "id": 1,
        "__class__": "quam.components.channels.InOutIQChannel",
    }

    bare_cfg = {
        "controllers": {},
        "elements": {},
        "pulses": {},
        "waveforms": {},
        "integration_weights": {},
    }
    cfg = bare_cfg.copy()
    expected_cfg = {
        "controllers": {
            "con1": {
                "analog_inputs": {
                    3: {"gain_db": 0, "shareable": False},
                    4: {"gain_db": 0, "shareable": False},
                },
                "analog_outputs": {
                    1: {"delay": 0, "shareable": False},
                    2: {"delay": 0, "shareable": False},
                },
            }
        },
        "elements": {
            "IQ1": {
                "intermediate_frequency": 100000000.0,
                "mixInputs": {
                    "I": ("con1", 1),
                    "Q": ("con1", 2),
                    "lo_frequency": 5000000000.0,
                    "mixer": "IQ1.mixer",
                },
                "operations": {},
                "outputs": {"out1": ("con1", 3), "out2": ("con1", 4)},
                "smearing": 0,
                "time_of_flight": 140,
            }
        },
        "pulses": {},
        "waveforms": {},
        "integration_weights": {},
    }
    readout_resonator.apply_to_config(cfg)
    assert cfg == expected_cfg

    cfg = bare_cfg.copy()
    readout_resonator._default_label = "res"
    readout_resonator.apply_to_config(cfg)
    expected_cfg["elements"]["res1"] = expected_cfg["elements"].pop("IQ1")
    expected_cfg["elements"]["res1"]["mixInputs"]["mixer"] = "IQ1.mixer"

    cfg = bare_cfg.copy()
    readout_resonator.id = "resonator_1"
    readout_resonator.apply_to_config(cfg)
    expected_cfg["elements"]["resonator_1"] = expected_cfg["elements"].pop("res1")
    expected_cfg["elements"]["resonator_1"]["mixInputs"]["mixer"] = "resonator_1.mixer"


def test_readout_resonator_with_readout():
    readout_resonator = InOutIQChannel(
        id=1,
        opx_output_I=("con1", 1),
        opx_output_Q=("con1", 2),
        opx_input_I=("con1", 3),
        opx_input_Q=("con1", 4),
        intermediate_frequency=100e6,
        frequency_converter_up=FrequencyConverter(
            mixer=Mixer(), local_oscillator=LocalOscillator(frequency=5e9)
        ),
    )
    readout_resonator.operations["readout"] = pulses.SquareReadoutPulse(
        amplitude=0.1, length=1000
    )

    d = readout_resonator.to_dict(include_defaults=False)
    assert d == {
        "frequency_converter_up": {
            "__class__": "quam.components.hardware.FrequencyConverter",
            "mixer": {"__class__": "quam.components.hardware.Mixer"},
            "local_oscillator": {
                "frequency": 5000000000.0,
                "__class__": "quam.components.hardware.LocalOscillator",
            },
        },
        "opx_output_I": ("con1", 1),
        "opx_output_Q": ("con1", 2),
        "opx_input_I": ("con1", 3),
        "opx_input_Q": ("con1", 4),
        "intermediate_frequency": 100000000.0,
        "id": 1,
        "operations": {
            "readout": {
                "__class__": "quam.components.pulses.SquareReadoutPulse",
                "amplitude": 0.1,
                "length": 1000,
            }
        },
        "__class__": "quam.components.channels.InOutIQChannel",
    }

    cfg = {
        "controllers": {},
        "elements": {},
        "pulses": {},
        "waveforms": {},
        "integration_weights": {},
        "digital_waveforms": {},
    }
    expected_cfg = {
        "controllers": {
            "con1": {
                "analog_inputs": {
                    3: {"gain_db": 0, "shareable": False},
                    4: {"gain_db": 0, "shareable": False},
                },
                "analog_outputs": {
                    1: {"delay": 0, "shareable": False},
                    2: {"delay": 0, "shareable": False},
                },
            }
        },
        "elements": {
            "IQ1": {
                "intermediate_frequency": 100000000.0,
                "mixInputs": {
                    "I": ("con1", 1),
                    "Q": ("con1", 2),
                    "lo_frequency": 5000000000.0,
                    "mixer": "IQ1.mixer",
                },
                "outputs": {"out1": ("con1", 3), "out2": ("con1", 4)},
                "smearing": 0,
                "time_of_flight": 140,
                "operations": {"readout": "IQ1.readout.pulse"},
            }
        },
        "pulses": {},
        "integration_weights": {},
        "waveforms": {},
        "digital_waveforms": {},
    }
    readout_resonator.apply_to_config(cfg)
    assert cfg == expected_cfg

    cfg = {
        "controllers": {},
        "elements": {},
        "pulses": {},
        "waveforms": {},
        "integration_weights": {},
        "digital_waveforms": {},
    }

    with pytest.raises(KeyError):
        readout_resonator.operations["readout"].apply_to_config(cfg)

    cfg["digital_waveforms"]["ON"] = {"samples": [(1, 0)]}
    readout_resonator.operations["readout"].apply_to_config(cfg)
    expected_cfg = {
        "controllers": {},
        "digital_waveforms": {"ON": {"samples": [(1, 0)]}},
        "elements": {},
        "integration_weights": {
            "IQ1.readout.iw1": {"cosine": [(1.0, 1000)], "sine": [(-0.0, 1000)]},
            "IQ1.readout.iw2": {"cosine": [(0.0, 1000)], "sine": [(1.0, 1000)]},
            "IQ1.readout.iw3": {"cosine": [(-0.0, 1000)], "sine": [(-1.0, 1000)]},
        },
        "pulses": {
            "IQ1.readout.pulse": {
                "digital_marker": "ON",
                "integration_weights": {
                    "iw1": "IQ1.readout.iw1",
                    "iw2": "IQ1.readout.iw2",
                    "iw3": "IQ1.readout.iw3",
                },
                "length": 1000,
                "operation": "measurement",
                "waveforms": {"I": "IQ1.readout.wf.I", "Q": "IQ1.readout.wf.Q"},
            }
        },
        "waveforms": {
            "IQ1.readout.wf.I": {"sample": 0.1, "type": "constant"},
            "IQ1.readout.wf.Q": {"sample": 0.0, "type": "constant"},
        },
    }
    assert cfg == expected_cfg


def test_channel_measure(mocker):
    readout_resonator = InOutIQChannel(
        id=1,
        opx_output_I=("con1", 1),
        opx_output_Q=("con1", 2),
        opx_input_I=("con1", 3),
        opx_input_Q=("con1", 4),
        intermediate_frequency=100e6,
        frequency_converter_up=FrequencyConverter(
            mixer=Mixer(), local_oscillator=LocalOscillator(frequency=5e9)
        ),
    )
    readout_resonator.operations["readout"] = pulses.SquareReadoutPulse(
        amplitude=0.1, length=1000
    )

    mocker.patch("quam.components.channels.declare", return_value=1)
    mocker.patch("quam.components.channels.measure", return_value=1)
    result = readout_resonator.measure("readout")
    assert result == (1, 1)


def test_empty_in_out_IQ_channel_ports():
    readout_resonator = InOutIQChannel(
        frequency_converter_up=FrequencyConverter(
            mixer=Mixer(), local_oscillator=LocalOscillator()
        ),
        opx_output_I=OPXPlusAnalogOutputPort("con1", 1),
        opx_output_Q=OPXPlusAnalogOutputPort("con1", 2),
        opx_input_I=OPXPlusAnalogInputPort("con1", 3),
        opx_input_Q=OPXPlusAnalogInputPort("con1", 4),
        intermediate_frequency=100e6,
    )

    assert isinstance(readout_resonator.frequency_converter_up.mixer, Mixer)

    assert isinstance(readout_resonator.local_oscillator, LocalOscillator)
    assert (
        readout_resonator.frequency_converter_up.local_oscillator
        is readout_resonator.local_oscillator
    )

    mixer = readout_resonator.frequency_converter_up.mixer
    assert mixer.intermediate_frequency == 100e6

    assert mixer.local_oscillator_frequency is None
    readout_resonator.local_oscillator.frequency = 5e9
    assert mixer.local_oscillator_frequency == 5e9

    with pytest.raises(AttributeError):
        mixer.name
    assert readout_resonator.id is None
    with pytest.raises(AttributeError):
        readout_resonator.name

    readout_resonator.id = 1

    d = readout_resonator.to_dict(include_defaults=False)
    assert d == {
        "frequency_converter_up": {
            "__class__": "quam.components.hardware.FrequencyConverter",
            "mixer": {"__class__": "quam.components.hardware.Mixer"},
            "local_oscillator": {
                "frequency": 5000000000.0,
                "__class__": "quam.components.hardware.LocalOscillator",
            },
        },
        "opx_output_I": {
            "controller_id": "con1",
            "port_id": 1,
            "__class__": "quam.components.ports.analog_outputs.OPXPlusAnalogOutputPort",
        },
        "opx_output_Q": {
            "controller_id": "con1",
            "port_id": 2,
            "__class__": "quam.components.ports.analog_outputs.OPXPlusAnalogOutputPort",
        },
        "opx_input_I": {
            "controller_id": "con1",
            "port_id": 3,
            "__class__": "quam.components.ports.analog_inputs.OPXPlusAnalogInputPort",
        },
        "opx_input_Q": {
            "controller_id": "con1",
            "port_id": 4,
            "__class__": "quam.components.ports.analog_inputs.OPXPlusAnalogInputPort",
        },
        "intermediate_frequency": 100000000.0,
        "id": 1,
        "__class__": "quam.components.channels.InOutIQChannel",
    }

    bare_cfg = {
        "controllers": {},
        "elements": {},
        "pulses": {},
        "waveforms": {},
        "integration_weights": {},
    }
    cfg = bare_cfg.copy()
    expected_cfg = {
        "controllers": {
            "con1": {
                "analog_inputs": {
                    3: {"gain_db": 0, "shareable": False},
                    4: {"gain_db": 0, "shareable": False},
                },
                "analog_outputs": {
                    1: {"delay": 0, "shareable": False},
                    2: {"delay": 0, "shareable": False},
                },
            }
        },
        "elements": {
            "IQ1": {
                "intermediate_frequency": 100000000.0,
                "mixInputs": {
                    "I": ("con1", 1),
                    "Q": ("con1", 2),
                    "lo_frequency": 5000000000.0,
                    "mixer": "IQ1.mixer",
                },
                "operations": {},
                "outputs": {"out1": ("con1", 3), "out2": ("con1", 4)},
                "smearing": 0,
                "time_of_flight": 140,
            }
        },
        "pulses": {},
        "waveforms": {},
        "integration_weights": {},
    }
    readout_resonator.apply_to_config(cfg)
    readout_resonator.opx_output_I.apply_to_config(cfg)
    readout_resonator.opx_output_Q.apply_to_config(cfg)
    readout_resonator.opx_input_I.apply_to_config(cfg)
    readout_resonator.opx_input_Q.apply_to_config(cfg)
    assert cfg == expected_cfg

    cfg = bare_cfg.copy()
    readout_resonator._default_label = "res"
    readout_resonator.apply_to_config(cfg)
    expected_cfg["elements"]["res1"] = expected_cfg["elements"].pop("IQ1")
    expected_cfg["elements"]["res1"]["mixInputs"]["mixer"] = "IQ1.mixer"

    cfg = bare_cfg.copy()
    readout_resonator.id = "resonator_1"
    readout_resonator.apply_to_config(cfg)
    expected_cfg["elements"]["resonator_1"] = expected_cfg["elements"].pop("res1")
    expected_cfg["elements"]["resonator_1"]["mixInputs"]["mixer"] = "resonator_1.mixer"



from quam.components.channels import Channel


def test_channel_frame_rotation_2pi(mocker):
    mocker.patch("quam.components.channels.frame_rotation_2pi")

    channel = Channel(id="channel")

    channel.frame_rotation_2pi(0.5)

    from quam.components.channels import frame_rotation_2pi

    frame_rotation_2pi.assert_called_once_with(0.5, "channel")


def test_channel_update_frequency(mocker):
    mocker.patch("quam.components.channels.update_frequency")

    channel = Channel(id="channel")

    channel.update_frequency(100e6)

    from quam.components.channels import update_frequency

    update_frequency.assert_called_once_with("channel", 100e6, "Hz", False)



from __future__ import annotations

import json
import os
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Sequence, Tuple, Union, List

from quam.config import get_quam_config
from .base import AbstractSerialiser

if TYPE_CHECKING:
    from quam.core import QuamRoot  # Assuming quam.core exists


def convert_int_keys(obj: Any) -> Any:
    """
    JSON object hook to convert dictionary keys that are strings
    representing integers back into integers during loading.

    Args:
        obj: The object being processed by the JSON decoder.

    Returns:
        The processed object, with relevant keys converted to integers.
    """
    if not isinstance(obj, dict):
        return obj

    new_obj = {}
    for key, value in obj.items():
        try:
            # Attempt to convert key to int if it represents a digit
            int_key = int(key)
            new_obj[int_key] = value
        except ValueError:
            # Keep the original string key if conversion fails
            new_obj[key] = value
    return new_obj


class JSONSerialiser(AbstractSerialiser):
    """
    Serialiser for QUAM objects to JSON files, allowing for splitting
    content across multiple files.

    Attributes:
        default_filename (str): Default filename if saving all content to one file.
        default_foldername (str): Default folder name if splitting content.
        content_mapping (Dict[str, str]): Defines how to split QUAM
            object content into different files. Keys are component names (top-level
            keys in the QUAM object's dictionary representation), and values are the
            relative filenames they should be saved to. If empty, saves to a single file
        include_defaults (bool): Whether to include default values in the
            serialised output.
        state_path (Optional[Path]): A specific path set during initialisation
            to be used as the default save/load location, overriding environment
            variables and configuration files.
    """

    default_filename: str = "state.json"
    default_foldername: str = "quam_state"
    content_mapping: Dict[str, str] = {}  # Expected final format: component -> filename

    def _resolve_include_defaults(self) -> bool:
        """
        Resolves the include_defaults setting using a priority chain.

        Priority (highest to lowest):
        1. Explicit parameter passed to save() (handled in save() method)
        2. Instance value (self.include_defaults)
        3. Config setting (quam.config.serialization.include_defaults)
        4. Fallback to True (default behavior)

        Returns:
            bool: Whether to include default values in serialization.
        """
        if self.include_defaults is not None:
            return self.include_defaults

        config = get_quam_config()
        if config.serialization is not None:
            return config.serialization.include_defaults
        return True

    @staticmethod
    def _validate_and_convert_content_mapping(
        mapping: Optional[Dict],
    ) -> Dict[str, str]:
        """
        Validates the content_mapping format and converts the old format
        (filename -> components) to the new format (component -> filename)
        if necessary, issuing a warning.

        Args:
            mapping: The content mapping dictionary to validate/convert.

        Returns:
            The validated content mapping in the new format (component -> filename).

        Raises:
            TypeError: If the mapping format is invalid (e.g., mixed value types).
        """
        if mapping is None:
            return {}

        if not isinstance(mapping, dict):
            raise TypeError(
                f"content_mapping must be a dictionary, got {type(mapping)}"
            )

        if not mapping:
            return {}

        # Check the type of the first value to infer format
        first_value = next(iter(mapping.values()))
        is_old_format = isinstance(
            first_value, (list, tuple, set, Sequence)
        ) and not isinstance(first_value, str)
        is_new_format = isinstance(first_value, str)

        if is_old_format:
            # Handle Old Format
            new_mapping: Dict[str, str] = {}
            conflicts: Dict[str, List[str]] = {}

            for filename, components in mapping.items():
                if not isinstance(filename, str):
                    raise TypeError(
                        "Invalid key in old format content_mapping: Expected string "
                        f"filename, got {type(filename)} ({filename})"
                    )
                if not isinstance(
                    components, (list, tuple, set, Sequence)
                ) or isinstance(components, str):
                    # Check for mixed formats within the old format assumption
                    raise TypeError(
                        "Mixed value types detected in content_mapping. Assumed old"
                        " format (filename -> components) based on first value, but"
                        f" found non-sequence value '{components}' for key"
                        f" '{filename}'."
                    )

                for component in components:
                    if not isinstance(component, str):
                        raise TypeError(
                            "Invalid component name in old format content_mapping:"
                            f" Expected string, got {type(component)} ({component}) in"
                            f" list for file '{filename}'"
                        )
                    if component in new_mapping:
                        conflicts.setdefault(
                            component, [new_mapping[component]]
                        ).append(filename)
                    new_mapping[component] = filename

            # Issue warnings for conflicts
            if conflicts:
                conflict_details = "; ".join(
                    [
                        f"'{comp}': [{', '.join(map(repr, files))}]"
                        for comp, files in conflicts.items()
                    ]
                )
                warnings.warn(
                    "Component conflicts detected in old format content_mapping. "
                    f"Components assigned to multiple files: {conflict_details}. "
                    "Using the last assignment found.",
                    UserWarning,
                )

            # Issue the main conversion warning
            old_repr = repr(mapping)
            new_repr = repr(new_mapping)
            warnings.warn(
                "Detected deprecated content_mapping format (filename -> components"
                " list).\nAutomatically converted to the new format (component ->"
                f" filename).\nOld mapping: {old_repr}\nConverted to:"
                f" {new_repr}\nPlease update your code to use the new format for future"
                " compatibility.",
                DeprecationWarning,
                stacklevel=3,  # Point warning towards the caller of __init__ or save
            )
            return new_mapping

        elif is_new_format:
            # Handle New Format
            # Validate that all values are strings
            for key, value in mapping.items():
                if not isinstance(key, str):
                    raise TypeError(
                        "Invalid key in new format content_mapping: Expected string"
                        f" component name, got {type(key)} ({key})"
                    )
                if not isinstance(value, str):
                    # Check for mixed formats within the new format assumption
                    raise TypeError(
                        "Mixed value types detected in content_mapping. Assumed new"
                        " format (component -> filename) based on first value, but"
                        f" found non-string value '{value}' for key '{key}'."
                    )
            return mapping  # Already in the correct format
        else:
            # Handle Invalid Format
            raise TypeError(
                "Invalid format for content_mapping. Values must be either all strings"
                " (component -> filename) or all sequences (filename -> components),"
                f" but found type {type(first_value)} for the first value."
            )

    def __init__(
        self,
        content_mapping: Optional[Dict] = None,  # Accept Dict initially for validation
        include_defaults: Optional[bool] = None,
        state_path: Optional[Union[str, Path]] = None,
    ):
        """
        Initialises the JSONSerialiser.

        Args:
            content_mapping: A specific content mapping to use. Can be in the old
                (filename->components) or new (component->filename) format. If old
                format is detected, a warning is issued and it's converted.
                If None, uses the class default.
            include_defaults: Whether to include fields set to their default
                values in the output. If None, follows priority chain:
                1. Config setting (quam.config.serialization.include_defaults)
                2. Fallback to True (default behavior)
            state_path: An optional default path for saving/loading state. If provided,
                this path takes precedence over environment variables or configuration
                files when determining the default save/load location.
        """
        initial_mapping = (
            content_mapping
            if content_mapping is not None
            else self.__class__.content_mapping
        )
        # Validate and potentially convert the mapping
        self.content_mapping = self._validate_and_convert_content_mapping(
            initial_mapping
        )

        self.include_defaults = include_defaults
        # Store the state_path, resolving it if provided
        self.state_path: Optional[Path] = (
            Path(state_path).resolve() if state_path else None
        )

    def _save_dict_to_json(self, contents: Dict[str, Any], filepath: Path):
        """
        Saves a dictionary to a specified JSON file.

        Args:
            contents: The dictionary to save.
            filepath: The exact path to the JSON file.
        """
        # Ensure parent directory exists just before writing
        filepath.parent.mkdir(parents=True, exist_ok=True)
        with filepath.open("w", encoding="utf-8") as f:
            json.dump(contents, fp=f, indent=4, ensure_ascii=False)

    def _save_split_content(
        self,
        full_contents: Dict[str, Any],
        folder: Path,
        content_mapping: Dict[str, str],  # Expects new format here
    ):
        """
        Saves dictionary content split across multiple files in a folder
        based on the content_mapping (component -> filename format).

        Args:
            full_contents: The complete dictionary of the QUAM object.
            folder: The target directory to save files into.
            content_mapping: Dictionary mapping component names (keys) to
                relative filenames (values). Old format (filename -> components)
                is not supported in this method.
        """
        remaining_contents = full_contents.copy()
        # Stores filepath -> content dict
        files_to_save: Dict[Path, Dict[str, Any]] = {}

        # Iterate through components and assign them to files based on mapping
        mapped_keys = set()
        for component_key, filename in content_mapping.items():
            if component_key in remaining_contents:
                # Resolve the filepath
                component_file = Path(filename)
                if component_file.is_absolute():
                    warnings.warn(
                        f"Absolute path '{filename}' in content_mapping is ignored. "
                        f"Using filename part only relative to '{folder}'.",
                        UserWarning,
                    )
                    component_filepath = folder / component_file.name
                else:
                    component_filepath = folder / component_file

                # Add component to the dictionary for this file
                if component_filepath not in files_to_save:
                    files_to_save[component_filepath] = {}
                files_to_save[component_filepath][component_key] = (
                    remaining_contents.pop(component_key)
                )
                mapped_keys.add(component_key)
            else:
                warnings.warn(
                    f"Key '{component_key}' specified in content_mapping was not found "
                    "in the QUAM object's data",
                    UserWarning,
                )

        # Save the collected components to their respective files
        for filepath, file_contents in files_to_save.items():
            self._save_dict_to_json(file_contents, filepath)

        # Save any remaining contents to the default file
        if remaining_contents:
            default_filepath = folder / self.default_filename
            self._save_dict_to_json(remaining_contents, default_filepath)

    def save(
        self,
        quam_obj: QuamRoot,
        path: Optional[Union[Path, str]] = None,
        content_mapping: Optional[Dict] = None,  # Accept Dict initially
        include_defaults: Optional[bool] = None,
        ignore: Optional[Sequence[str]] = None,
    ):
        """
        Saves a QuamRoot object to JSON file(s).

        Determines the save location and format (single file vs. split) based
        on the provided path and content_mapping arguments, falling back to
        instance/class defaults if arguments are None.

        Args:
            quam_obj: The QuamRoot object to serialise.
            path: The target file or folder path. If None, uses default path logic
                  (checking instance `state_path`, env var, config).
            content_mapping: Overrides the instance's content_mapping for this save.
                Can be in old or new format. If old format is provided, a warning is
                issued and it's converted internally.
            include_defaults: Overrides the instance's include_defaults for this save.
            ignore: A sequence of top-level keys in the QUAM object to exclude
                    from the saved output.
        """
        # Validate and convert the provided content_mapping, or use the instance's
        # (already validated) one
        if content_mapping is not None:
            current_content_mapping = self._validate_and_convert_content_mapping(
                content_mapping
            )
        else:
            current_content_mapping = self.content_mapping  # Already validated in init

        # Resolve include_defaults with priority:
        # parameter > instance > config > fallback
        if include_defaults is not None:
            current_include_defaults = include_defaults
        else:
            current_include_defaults = self._resolve_include_defaults()

        if path is None:
            path = self._get_state_path()
        path = Path(path)

        # Get the dictionary representation of the object
        contents_dict = quam_obj.to_dict(include_defaults=current_include_defaults)

        # Apply ignore filter directly to the source dictionary before saving
        # This modification is temporary for the save operation.
        effective_contents = contents_dict.copy()
        if ignore:
            current_content_mapping = current_content_mapping.copy()
            for key in ignore:
                effective_contents.pop(key, None)  # Modify the copy
                current_content_mapping.pop(key, None)

        if path.suffix == ".json":
            # Target is a json file, save content to it
            self._save_dict_to_json(effective_contents, path)
        elif not path.suffix:
            # Target is a directory, use split logic
            self._save_split_content(
                full_contents=effective_contents,
                folder=path,
                content_mapping=current_content_mapping,  # Pass validated mapping
            )
        else:
            raise ValueError(
                f"Cannot save QUAM: Unsupported path suffix '{path.suffix}'"
            )

    def _get_state_path(self) -> Path:
        """
        Determines the default path for saving/loading state.

        Resolution order:
        1. `self.state_path` (if set during `__init__`).
        2. `QUAM_STATE_PATH` environment variable.
        3. `state_path` from QUAM configuration (via `get_quam_config`).
        4. Fallback to `default_foldername` or `default_filename` in the current
           directory.

        Returns:
            The default Path object, resolved to an absolute path.
        """
        # 1. Check instance path first
        if self.state_path is not None:
            return self.state_path.resolve()

        # 2. Check environment variable
        env_path = os.environ.get("QUAM_STATE_PATH")
        if env_path:
            return Path(env_path).resolve()

        # 3. Check configuration file
        try:
            cfg = get_quam_config()
            if cfg and cfg.state_path is not None:
                return Path(cfg.state_path).resolve()

        except (AttributeError, FileNotFoundError):  # Catch potential errors
            warnings.warn(
                "Could not determine state path from QUAM configuration. "
                "Falling back to environment or default.",
                UserWarning,
            )

        # 4. No path found - Fallback to saving in current directory
        # Decide on using the folder or single file default based on content_mapping
        # Use the mapping already validated/converted in __init__
        if self.content_mapping:
            default_path = Path(self.default_foldername)
            warnings.warn(
                "No state path found via init, environment, or config. Defaulting "
                f"to folder '{default_path}' in the current directory because "
                "content_mapping is defined.",
                UserWarning,
            )
        else:
            default_path = Path(self.default_filename)
            warnings.warn(
                "No state path found via init, environment, or config. Defaulting "
                f"to file '{default_path}' in the current directory.",
                UserWarning,
            )
        return default_path.resolve()

    def _load_from_file(self, filepath: Path) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Loads dictionary content from a single JSON file.

        Args:
            filepath: The exact path to the JSON file.

        Returns:
            Tuple[Dict[str, Any], Dict[str, Any]]: A tuple containing the loaded
                dictionary content and metadata dictionary including inferred
                'content_mapping' (component -> filename), 'default_filename',
                and 'default_foldername'.

        Raises:
            TypeError: If the filepath does not have a .json suffix.
            json.JSONDecodeError: If the file content is invalid JSON.
            IOError: If the file cannot be read.
        """
        if filepath.suffix.lower() != ".json":
            raise TypeError(f"File {filepath} is not a JSON file.")

        try:
            with filepath.open("r", encoding="utf-8") as f:
                contents = json.load(f, object_hook=convert_int_keys)
            if not isinstance(contents, dict):
                raise TypeError(
                    f"File {filepath} does not contain a valid JSON dictionary.",
                )

            metadata = {
                "content_mapping": {},
                "default_filename": filepath.name,
                "default_foldername": None,
            }
            return contents, metadata
        except json.JSONDecodeError as e:
            raise json.JSONDecodeError(
                f"Error decoding JSON from {filepath}: {e.msg}", e.doc, e.pos
            ) from e
        except IOError as e:
            raise IOError(f"Error reading file {filepath}: {e}") from e
        except Exception as e:  # Catch unexpected errors during loading
            raise RuntimeError(
                f"An unexpected error occurred while loading {filepath}: {e}"
            ) from e

    def _load_from_directory(
        self, dirpath: Path
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Loads and merges content from all .json files in a directory and its
        subdirectories. Infers the content mapping (component -> filename).
        Skips hidden directories (those starting with a dot).

        Args:
            dirpath: The path to the directory to load from.

        Returns:
            Tuple[Dict[str, Any], Dict[str, Any]]: A tuple containing the merged
                dictionary content and metadata dictionary including inferred
                'content_mapping', 'default_filename', and 'default_foldername'.
        """
        contents: Dict[str, Any] = {}
        inferred_mapping: Dict[str, str] = {}  # component -> relative_filename
        metadata: Dict[str, Any] = {
            "content_mapping": inferred_mapping,
            "default_filename": None,
            "default_foldername": str(dirpath.resolve()),
        }

        # Find all JSON files, excluding those in hidden directories (starting with dot)
        found_files = []
        for json_file in dirpath.rglob("*.json"):
            # Check if any part of the path contains a directory starting with dot
            relative_path = json_file.relative_to(dirpath)
            if any(part.startswith(".") for part in relative_path.parts[:-1]):
                # Skip files in hidden directories
                continue
            found_files.append(json_file)

        if not found_files:
            warnings.warn(f"No JSON files found in directory {dirpath}", UserWarning)
            return contents, metadata

        processed_files_count = 0
        for file in found_files:
            try:
                file_contents, _ = self._load_from_file(
                    file
                )  # Metadata from single file load is not needed here
                if not file_contents:  # Skip empty files
                    warnings.warn(
                        f"Skipping empty or invalid JSON file: {file}", UserWarning
                    )
                    continue

                relative_filepath = file.relative_to(dirpath).as_posix()
                processed_files_count += 1

                # Check for key conflicts before updating contents
                conflicts = contents.keys() & file_contents.keys()
                if conflicts:
                    conflict_details = {
                        key: inferred_mapping[key]
                        for key in conflicts
                        if key in inferred_mapping
                    }
                    warnings.warn(
                        f"Key conflicts detected: Components {list(conflicts)} found in"
                        f" '{relative_filepath}' overwrite existing definitions from"
                        f" files {conflict_details}. Using definition from"
                        f" '{relative_filepath}'.",
                        UserWarning,
                    )
                contents.update(file_contents)

                # Update inferred mapping: component -> relative_filepath
                for key in file_contents.keys():
                    # Overwrite mapping on conflict
                    inferred_mapping[key] = relative_filepath

                # Check if this file is the default file at the root level
                if file.name == self.default_filename and file.parent == dirpath:
                    metadata["default_filename"] = file.name

            except (
                json.JSONDecodeError,
                IOError,
                TypeError,
                RuntimeError,
            ) as e:
                warnings.warn(f"Skipping file {file} due to error: {e}", UserWarning)

        if processed_files_count == 0 and found_files:
            warnings.warn(
                f"Found {len(found_files)} JSON files in {dirpath}, but none contained"
                " valid data.",
                UserWarning,
            )

        # Filter out inferred_mapping keys that point to the default file
        inferred_mapping = {
            key: value
            for key, value in inferred_mapping.items()
            if value != self.default_filename
        }

        metadata["content_mapping"] = inferred_mapping
        return contents, metadata

    def load(
        self,
        path: Optional[Union[Path, str]] = None,
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        """
        Loads a QuamRoot object's dictionary representation from JSON file(s).

        Determines the load path using the provided `path` argument or falls back
        to default logic (instance `state_path`, env var, config). Loads from a
        single file or merges from a directory (recursively).

        Args:
            path: The path to load from (file or directory). If None, uses the
                  default state path logic via `_get_state_path()`.

        Returns:
            Tuple[Dict[str, Any], Dict[str, Any]]: A tuple containing the dictionary
                representation of the loaded QUAM object and metadata dictionary
                including inferred 'content_mapping' (component -> filename),
                'default_filename', and 'default_foldername'.

        Raises:
            FileNotFoundError: If the resolved path does not exist.
            ValueError: If the resolved path is neither a file nor a directory.
        """
        load_path: Path
        if path is None:
            load_path = self._get_state_path()
        else:
            load_path = Path(path).resolve()

        if not load_path.exists():
            raise FileNotFoundError(f"Path {load_path} not found, cannot load JSON.")

        contents: Dict[str, Any]
        metadata: Dict[str, Any]

        if load_path.is_file():
            contents, metadata = self._load_from_file(load_path)
        elif load_path.is_dir():
            contents, metadata = self._load_from_directory(load_path)
        else:
            raise ValueError(f"Path {load_path} is neither a valid file nor directory.")

        # Update the instance's content_mapping *only if* it wasn't explicitly set
        # during initialization or via the save method argument override.
        is_default_mapping = self.content_mapping == self.__class__.content_mapping
        if is_default_mapping and "content_mapping" in metadata:
            self.content_mapping = metadata["content_mapping"]

        return contents, metadata



from copy import deepcopy
from quam.components import BasicQuam, SingleChannel


def test_instantiate_basic_quam_empty():
    quam = BasicQuam()
    assert quam.channels == {}
    assert quam.octaves == {}

    d = quam.to_dict(include_defaults=False)
    assert d == {
        "__class__": "quam.components.basic_quam.BasicQuam",
    }


def test_instantiate_basic_quam_with_channel():
    quam = BasicQuam(channels={"single": SingleChannel(opx_output=("con1", 1))})
    d = quam.to_dict(include_defaults=False)
    assert d == {
        "__class__": "quam.components.basic_quam.BasicQuam",
        "channels": {
            "single": {
                "__class__": "quam.components.channels.SingleChannel",
                "opx_output": ("con1", 1),
            },
        },
    }


def test_generate_basic_quam_config(qua_config):
    quam = BasicQuam()
    cfg = quam.generate_config()
    assert cfg == qua_config

    quam.channels = {"single": SingleChannel(opx_output=("con1", 1))}
    cfg = quam.generate_config()

    quam.channels["single"].apply_to_config(qua_config)
    qua_config["controllers"]["con1"]["analog_outputs"][1]["offset"] = 0.0

    assert cfg == qua_config



import importlib
import warnings
from inspect import isclass
from typing import Any, Union

from quam.utils import string_reference

from typeguard import TypeCheckError, check_type

__all__ = ["get_full_class_path", "validate_obj_type", "get_class_from_path"]


def get_full_class_path(cls_or_obj: Union[type, object]) -> str:
    """Returns the full path of a class or object, including the module name.

    Example:
    ```
    from quam.components import Mixer
    assert get_full_class_path(Mixer) == "quam.components.hardware.Mixer"
    ```

    Args:
        cls_or_obj: The class or object to get the path of.

    Returns:
        The full path of the class or object. Generally this is of the form
        "module_name.class_name".

    Warnings:
        If the module name cannot be determined, a warning is raised.
    """
    if isclass(cls_or_obj):
        class_name = cls_or_obj.__qualname__
    else:
        class_name = cls_or_obj.__class__.__qualname__

    module_name = cls_or_obj.__module__
    if module_name == "__main__" or module_name is None:
        warnings.warn(
            f"Could not determine the module of {class_name}, this may cause issues"
            " when trying to load QUAM from a file. Please ensure that all QUAM"
            " classes are defined in a Python module"
        )
        return class_name
    else:
        return f"{module_name}.{class_name}"


def validate_obj_type(
    elem: Any, required_type: type, allow_none: bool = True, str_repr: str = ""
) -> None:
    """Validate whether the object is an instance of the correct type

    References (strings starting with "#") are not checked.
    None is always allowed.

    Args:
        elem: The object to validate the type of.
        required_type: The required type of the object.
        allow_none: Whether None is allowed as a value even if it's the wrong type.
        str_repr: A string representation of the object, used for error messages.

    Returns:
        None

    Raises:
        TypeError if the type of the attribute is not the required type
    """
    # Do not check type if the value is a reference
    if string_reference.is_reference(elem):
        return
    if elem is None and allow_none:
        return
    try:
        check_type(elem, required_type)
    except TypeCheckError as e:
        if elem is None:
            raise TypeError(
                f"None is not allowed for required attribute {str_repr}"
            ) from e
        else:
            raise TypeError(
                "Wrong object type found during validation.\n"
                f"Path: {str_repr}\n"
                f"Required type: {required_type}\n"
                f"Actual type: {type(elem)}\n"
                f"value of actual type: {elem}"
            ) from e


def get_class_from_path(class_str) -> type:
    """Extract the class from a class path.

    Example:
        ```
        from quam.components import Mixer
        assert get_class_from_path("quam.components.hardware.Mixer") == Mixer
        ```

    Args:
        class_str: The class path, e.g. "quam.components.hardware.Mixer"

    Returns:
        Class object corresponding to the class path.
    """
    try:
        module_path, class_name = class_str.rsplit(".", 1)
    except ValueError as e:
        raise ValueError(
            "Could not extract module and class name from class path, be sure that the "
            "class path is of the form '{module_name}.{class_name}'. "
            f"class_str: '{class_str}'"
        ) from e
    module = importlib.import_module(module_path)
    quam_class = getattr(module, class_name)
    return quam_class



from quam.components.channels import InSingleChannel, TimeTaggingAddon
from quam.core.quam_classes import QuamRoot, quam_dataclass


@quam_dataclass
class SingleChannelQuam(QuamRoot):
    channel: InSingleChannel


def test_time_tagging_cfg():
    channel = InSingleChannel(id="channel", opx_input=("con1", 1))
    channel.time_tagging = TimeTaggingAddon()

    machine = SingleChannelQuam(channel=channel)
    cfg = machine.generate_config()

    assert "timeTaggingParameters" in cfg["elements"]["channel"]
    assert cfg["elements"]["channel"]["timeTaggingParameters"] == {
        "signalThreshold": 800,
        "signalPolarity": "below",
        "derivativeThreshold": 300,
        "derivativePolarity": "below",
    }


def test_time_tagging_cfg_disabled():
    channel = InSingleChannel(id="channel", opx_input=("con1", 1))
    channel.time_tagging = TimeTaggingAddon(enabled=False)

    machine = SingleChannelQuam(channel=channel)
    cfg = machine.generate_config()
    assert "timeTaggingParameters" not in cfg["elements"]["channel"]


def test_time_tagging_cfg_custom_thresholds():
    channel = InSingleChannel(id="channel", opx_input=("con1", 1))
    channel.time_tagging = TimeTaggingAddon(
        signal_threshold=0.2, derivative_threshold=0.1
    )

    machine = SingleChannelQuam(channel=channel)
    cfg = machine.generate_config()
    assert "timeTaggingParameters" in cfg["elements"]["channel"]
    assert cfg["elements"]["channel"]["timeTaggingParameters"] == {
        "signalThreshold": int(0.2 * 4096),
        "signalPolarity": "below",
        "derivativeThreshold": int(0.1 * 4096),
        "derivativePolarity": "below",
    }



import pytest
from quam.core.quam_classes import quam_dataclass
from quam.components import Qubit
from quam.components.macro import QubitMacro


@quam_dataclass
class TestMacro(QubitMacro):
    """Simple macro class for testing purposes"""

    def apply(self, *args, **kwargs):
        # Return inputs to verify they were passed correctly
        return (self.qubit, args, kwargs)


@pytest.fixture
def test_qubit():
    """Fixture providing a qubit with common test macros"""
    qubit = Qubit(id="test_qubit")

    # Add some common macros
    qubit.macros["x_gate"] = TestMacro()
    qubit.macros["test_op"] = TestMacro()

    return qubit



from .dataclass import *
from .pulse import *
from .reference_class import *
from .type_checking import *
from .general import *
from . import string_reference
from .config import *

__all__ = [
    *dataclass.__all__,
    *general.__all__,
    *pulse.__all__,
    *reference_class.__all__,
    *type_checking.__all__,
    "string_reference",
    *config.__all__,
]
